import pathlib
import datetime
from typing import Optional

import h5py
import numpy

from larch.io.xas_data_source import open_xas_source
from larch.io.xas_data_source import read_xas_source


def test_xas_data_source_spec(tmp_path):
    filename = tmp_path / "scans.spec"

    energy1 = numpy.linspace(7, 7.1, 10)
    energy2 = numpy.linspace(7, 7.1, 20)
    mu1 = numpy.random.uniform(low=-1, high=2, size=10)
    mu2 = numpy.random.uniform(low=-1, high=2, size=20)

    _save_spec_scan(filename, "escan", energy=energy1, mu=mu1)
    _save_spec_scan(
        filename,
        "ascan",
        samx=numpy.linspace(1, 2, 10),
        diode1=numpy.random.uniform(size=10),
    )
    _save_spec_scan(filename, "escan", energy=energy2, mu=mu2)

    source = open_xas_source(str(filename), title_regex_pattern="escan")
    assert source.TYPE == "SPEC"
    assert source.get_source_info() == f"SPEC: {filename}"
    assert source.get_scan_names() == ["1.1", "3.1"]

    scan1 = source.get_scan("1.1")
    assert scan1.labels == ["energy", "mu"]
    numpy.testing.assert_array_equal(energy1, scan1.data[0])
    numpy.testing.assert_array_equal(mu1, scan1.data[1])

    scan2 = source.get_scan("3.1")
    assert scan2.labels == ["energy", "mu"]
    numpy.testing.assert_array_equal(energy2, scan2.data[0])
    numpy.testing.assert_array_equal(mu2, scan2.data[1])

    scan1 = read_xas_source(str(filename), "1.1")
    numpy.testing.assert_array_equal(energy1, scan1.energy)
    numpy.testing.assert_array_equal(mu1, scan1.mu)

    scan2 = read_xas_source(str(filename), "3.1")
    numpy.testing.assert_array_equal(energy2, scan2.energy)
    numpy.testing.assert_array_equal(mu2, scan2.mu)


def test_xas_data_source_esrf(tmp_path):
    filename = tmp_path / "dataset.h5"

    energy1 = numpy.linspace(7, 7.1, 10)
    energy2 = numpy.linspace(7, 7.1, 20)
    mu1 = numpy.random.uniform(low=-1, high=2, size=10)
    mu2 = numpy.random.uniform(low=-1, high=2, size=20)

    _save_nexus_scan(filename, "escan", facility="esrf", energy=energy1, mu=mu1)
    _save_nexus_scan(
        filename,
        "ascan",
        facility="esrf",
        samx=numpy.linspace(1, 2, 10),
        diode1=numpy.random.uniform(size=10),
    )
    _save_nexus_scan(filename, "escan", facility="esrf", energy=energy2, mu=mu2)

    source = open_xas_source(str(filename), title_regex_pattern="escan")
    assert source.TYPE == "HDF5-NEXUS-ESRF"
    assert source.get_source_info() == f"HDF5: {filename}"
    assert source.get_scan_names() == ["1.1", "3.1"]

    scan1 = source.get_scan("1.1")
    assert scan1.labels == ["energy", "mu"]
    numpy.testing.assert_array_equal(energy1, scan1.data[0])
    numpy.testing.assert_array_equal(mu1, scan1.data[1])

    scan2 = source.get_scan("3.1")
    assert scan2.labels == ["energy", "mu"]
    numpy.testing.assert_array_equal(energy2, scan2.data[0])
    numpy.testing.assert_array_equal(mu2, scan2.data[1])

    scan1 = read_xas_source(str(filename), "1.1")
    numpy.testing.assert_array_equal(energy1, scan1.energy)
    numpy.testing.assert_array_equal(mu1, scan1.mu)

    scan2 = read_xas_source(str(filename), "3.1")
    numpy.testing.assert_array_equal(energy2, scan2.energy)
    numpy.testing.assert_array_equal(mu2, scan2.mu)


def test_xas_data_source_soleil(tmp_path):
    filename = tmp_path / "dataset.nxs"

    energy1 = numpy.linspace(7, 7.1, 10)
    energy2 = numpy.linspace(7, 7.1, 20)
    mu1 = numpy.random.uniform(low=-1, high=2, size=10)
    mu2 = numpy.random.uniform(low=-1, high=2, size=20)

    _save_nexus_scan(filename, "escan", facility="soleil", energy=energy1, mu=mu1)
    _save_nexus_scan(
        filename,
        "ascan",
        facility="soleil",
        samx=numpy.linspace(1, 2, 10),
        diode1=numpy.random.uniform(size=10),
    )
    _save_nexus_scan(filename, "escan", facility="soleil", energy=energy2, mu=mu2)

    source = open_xas_source(str(filename), title_regex_pattern="escan")
    assert source.TYPE == "HDF5-NEXUS-SOLEIL"
    assert source.get_source_info() == f"HDF5: {filename}"
    assert source.get_scan_names() == ["exp1", "exp3"]

    scan1 = source.get_scan("exp1")
    assert scan1.labels == ["energy", "mu"]
    numpy.testing.assert_array_equal(energy1, scan1.data[0])
    numpy.testing.assert_array_equal(mu1, scan1.data[1])

    scan2 = source.get_scan("exp3")
    assert scan2.labels == ["energy", "mu"]
    numpy.testing.assert_array_equal(energy2, scan2.data[0])
    numpy.testing.assert_array_equal(mu2, scan2.data[1])

    scan1 = read_xas_source(str(filename), "exp1")
    numpy.testing.assert_array_equal(energy1, scan1.energy)
    numpy.testing.assert_array_equal(mu1, scan1.mu)

    scan2 = read_xas_source(str(filename), "exp3")
    numpy.testing.assert_array_equal(energy2, scan2.energy)
    numpy.testing.assert_array_equal(mu2, scan2.mu)


def test_xas_data_source_nexus(tmp_path):
    filename = tmp_path / "dataset.nxs"

    energy1 = numpy.linspace(7, 7.1, 10)
    energy2 = numpy.linspace(7, 7.1, 20)
    mu1 = numpy.random.uniform(low=-1, high=2, size=10)
    mu2 = numpy.random.uniform(low=-1, high=2, size=20)

    _save_nexus_scan(filename, "escan", energy=energy1, mu=mu1)
    _save_nexus_scan(
        filename,
        "ascan",
        samx=numpy.linspace(1, 2, 10),
        diode1=numpy.random.uniform(size=10),
    )
    _save_nexus_scan(filename, "escan", energy=energy2, mu=mu2)

    source = open_xas_source(str(filename), title_regex_pattern="escan")
    assert source.TYPE == "HDF5-NEXUS"
    assert source.get_source_info() == f"HDF5: {filename}"
    assert source.get_scan_names() == ["scan1", "scan3"]

    scan1 = source.get_scan("scan1")
    assert scan1.labels == ["energy", "mu"]
    numpy.testing.assert_array_equal(energy1, scan1.data[0])
    numpy.testing.assert_array_equal(mu1, scan1.data[1])

    scan2 = source.get_scan("scan3")
    assert scan2.labels == ["energy", "mu"]
    numpy.testing.assert_array_equal(energy2, scan2.data[0])
    numpy.testing.assert_array_equal(mu2, scan2.data[1])

    scan1 = read_xas_source(str(filename), "scan1")
    numpy.testing.assert_array_equal(energy1, scan1.energy)
    numpy.testing.assert_array_equal(mu1, scan1.mu)

    scan2 = read_xas_source(str(filename), "scan3")
    numpy.testing.assert_array_equal(energy2, scan2.energy)
    numpy.testing.assert_array_equal(mu2, scan2.mu)


def _save_spec_scan(filename: pathlib.Path, scan_title: str, **data) -> None:
    date = datetime.datetime.now().strftime("%a %b %d %H:%M:%S %Y")

    header = list()
    if filename.exists():
        with open(filename, "r") as f:
            scan_number = len([None for line in f if line.startswith("#S")]) + 1
    else:
        header += [f"#F {filename}", f"#D {date}"]
        scan_number = 1
    header += [
        "",
        f"#S {scan_number} {scan_title}",
        f"#D {date}",
        f"#N {len(data)}",
        f"#L {'  '.join(list(data))}",
    ]

    array = numpy.array(list(data.values())).T
    with open(filename, "ab") as f:
        numpy.savetxt(f, array, header="\n".join(header), comments="")


def _save_nexus_scan(
    filename: pathlib.Path, scan_title: str, facility: Optional[str] = None, **data
) -> None:
    with h5py.File(filename, "a") as nxroot:
        scan_names = list(nxroot)
        if not scan_names:
            scan_numbers = [0]
            nxroot.attrs["NX_class"] = "NXroot"
            if facility == "esrf":
                nxroot.attrs["creator"] = "Bliss"
        else:
            if facility == "esrf":
                scan_numbers = [int(float(name)) for name in scan_names]
            elif facility == "soleil":
                scan_numbers = [int(name.replace("exp", "")) for name in scan_names]
            else:
                scan_numbers = [int(name.replace("scan", "")) for name in scan_names]

        if facility == "esrf":
            scan_name = f"{max(scan_numbers)+1}.1"
            instrument_name = "instrument"
            source_name = "machine"
            source_info = {
                "type": "Synchrotron X-ray Source",
                "name": "ESRF",
            }
            scan_data_name = "measurement"
            scan_data_type = "NXcollection"
        elif facility == "soleil":
            scan_name = f"exp{max(scan_numbers)+1}"
            instrument_name = "PUMA"
            source_name = "ans-ca-machinestatus"
            source_info = {
                "type": "Synchrotron X-ray Source",
                "probe": "x-ray",
                "name": "SOLEIL",
            }
            scan_data_name = "scan_data"
            scan_data_type = "NXdata"
        else:
            scan_name = f"scan{max(scan_numbers)+1}"
            instrument_name = "instrument"
            source_name = None
            scan_data_name = None
            scan_data_type = None

        nxentry = nxroot.create_group(scan_name)
        nxentry.attrs["NX_class"] = "NXentry"
        nxentry["title"] = scan_title

        if facility == "esrf":
            nxentry["name"] = "ESRF-ID21"
            nxentry["name"].attrs["short_name"] = "ID21"

        nxinstrument = nxentry.create_group(instrument_name)
        nxinstrument.attrs["NX_class"] = "NXinstrument"

        if source_name:
            nxsource = nxinstrument.create_group(source_name)
            nxsource.attrs["NX_class"] = "NXsource"
            for name, value in source_info.items():
                nxsource[name] = value

        if scan_data_name:
            scan_data = nxentry.create_group(scan_data_name)
            scan_data.attrs["NX_class"] = scan_data_type

        for name, value in data.items():
            if name.startswith("sam"):
                nxpositioner = nxinstrument.create_group(name)
                nxpositioner.attrs["NX_class"] = "NXpositioner"
                dset = nxpositioner.create_dataset("value", data=value)
                dset.attrs["units"] = "mm"
            elif name == "energy":
                nxpositioner = nxinstrument.create_group(name)
                nxpositioner.attrs["NX_class"] = "NXpositioner"
                dset = nxpositioner.create_dataset("value", data=value)
                dset.attrs["units"] = "keV"
            else:
                nxdetector = nxinstrument.create_group(name)
                nxdetector.attrs["NX_class"] = "NXdetector"
                dset = nxdetector.create_dataset("data", data=value)
                dset.attrs["units"] = "keV"

            if scan_data_name:
                scan_data[name] = h5py.SoftLink(dset.name)



#!/usr/bin/env python
"""
merge groups, interpolating if necessary
"""
import os
import numpy as np
from larch import Group
from larch.math import interp, interp1d, index_of, remove_dups
from larch.utils.logging import getLogger

_logger = getLogger("larch.io.mergegroups")

def merge_groups(grouplist, master=None, xarray='energy', yarray='mu',
                 kind='cubic', trim=True, calc_yerr=True):
    """merge arrays from a list of groups.

    Arguments
    ---------
     grouplist   list of groups to merge
     master      group to use for common x arrary [None -> 1st group]
     xarray      name of x-array for merge ['energy']
     yarray      name of y-array for merge ['mu']
     kind        interpolation kind ['cubic']
     trim        whether to trim to the shortest energy range [True]
     calc_yerr   whether to use the variance in the input as yerr [True]

    Returns
    --------
     group with x-array and y-array containing merged data.

    """
    if master is None:
        master = grouplist[0]

    xout = remove_dups(getattr(master, xarray))
    xmins = [min(xout)]
    xmaxs = [max(xout)]
    yvals = []

    for g in grouplist:
        x = getattr(g, xarray)
        y = getattr(g, yarray)
        yvals.append(interp(x, y, xout, kind=kind))
        xmins.append(min(x))
        xmaxs.append(max(x))

    yvals = np.array(yvals)
    yave = yvals.mean(axis=0)
    ystd = yvals.std(axis=0)

    xout_increasing = len(np.where(np.diff(np.argsort(xout))!=1)[0]) == 0
    if trim and xout_increasing:
        xmin = min(xmins)
        xmax = min(xmaxs)
        ixmin = index_of(xout, xmin)
        ixmax = index_of(xout, xmax)
        xout = xout[ixmin:ixmax]
        yave = yave[ixmin:ixmax]
        ystd = ystd[ixmin:ixmax]

    grp = Group()
    setattr(grp, xarray, xout)
    setattr(grp, yarray, yave)
    setattr(grp, yarray + '_std', ystd)

    if kind == 'cubic' and xout_increasing:
        y0 = getattr(master, yarray)
        # if the derivative gets much worse, use linear interpolation
        if max(np.diff(yave)) > 50*max(np.diff(y0)):
            grp = merge_groups(grouplist, master=master, xarray=xarray,
                               yarray=yarray, trim=trim,
                               calc_yerr=calc_yerr, kind='linear')
    return grp

def imin(arr, debug=False):
    """index of minimum value"""
    _im = np.argmin(arr)
    if debug:
        _logger.debug("Check: {0} = {1}".format(np.min(arr), arr[_im]))
    return _im


def imax(arr, debug=False):
    """index of maximum value"""
    _im = np.argmax(arr)
    if debug:
        _logger.debug("Check: {0} = {1}".format(np.max(arr), arr[_im]))
    return _im


def lists_to_matrix(data, axis=None, **kws):
    """Convert two lists of 1D arrays to a 2D matrix

    Parameters
    ----------
    data : list of lists of 1D arrays
        [
            [x1, ... xN]
            [z1, ... zN]
        ]
    axis : None or array 1D, optional
        a reference axis used for the interpolation [None -> xdats[0]]
    **kws : optional
        keyword arguments for scipy.interpolate.interp1d()

    Returns
    -------
    axis, outmat : arrays
    """
    assert len(data) == 2, "'data' should be a list of two lists"
    xdats, zdats = data
    assert isinstance(xdats, list), "'xdats' not a list"
    assert isinstance(zdats, list), "'zdats' not a list"
    assert len(xdats) == len(zdats), "lists of data not of the same length"
    assert all(isinstance(z, np.ndarray) for z in zdats), "data in list must be arrays"
    if axis is None:
        axis = xdats[0]
    assert isinstance(axis, np.ndarray), "axis must be array"
    if all(z.size == axis.size for z in zdats):
        #: all same size
        return axis, np.array(zdats)
    else:
        #: interpolate
        outmat = np.zeros((len(zdats), axis.size))
        for idat, (xdat, zdat) in enumerate(zip(xdats, zdats)):
            fdat = interp1d(xdat, zdat, **kws)
            znew = fdat(axis)
            outmat[idat] = znew
        return axis, outmat


def curves_to_matrix(curves, axis=None, **kws):
    """Convert a list of curves to a 2D data matrix

    Parameters
    ----------
    curves : list of lists
        Curves format is the following:
        [
            [x1, y1, label1, info1],
            ...
            [xN, yN, labelN, infoN]
        ]
    axis : None or array 1D, optional
        a reference axis used for the interpolation [None -> curves[0][0]]
    **kws : optional
        keyword arguments for func:`scipy.interpolate.interp1d`

    Returns
    -------
    axis, outmat : arrays
    """
    assert isinstance(curves, list), "'curves' not a list"
    assert all(
        (isinstance(curve, list) and len(curve) == 4) for curve in curves
    ), "curves should be lists of four elements"
    if axis is None:
        axis = curves[0][0]
    assert isinstance(axis, np.ndarray), "axis must be array"
    outmat = np.zeros((len(curves), axis.size))
    for icurve, curve in enumerate(curves):
        assert len(curve) == 4, "wrong curve format, should contain four elements"
        x, y, label, info = curve
        try:
            assert isinstance(x, np.ndarray), "not array!"
            assert isinstance(y, np.ndarray), "not array!"
        except AssertionError:
            _logger.error(
                "[curve_to_matrix] Curve %d (%s) not containing arrays -> ADDING ZEROS",
                icurve,
                label,
            )
            continue
        if (x.size == axis.size) and (y.size == axis.size):
            #: all same length
            outmat[icurve] = y
        else:
            #: interpolate
            fdat = interp1d(x, y, **kws)
            ynew = fdat(axis)
            outmat[icurve] = ynew
            _logger.debug("[curve_to_matrix] Curve %d (%s) interpolated", icurve, label)
    return axis, outmat


def sum_arrays_1d(data, axis=None, **kws):
    """Sum list of 1D arrays or curves by interpolation on a reference axis

    Parameters
    ----------
    data : lists of lists
    data_fmt : str
        define data format
        - "curves" -> :func:`curves_to_matrix`
        - "lists" -> :func:`curves_to_matrix`

    Returns
    -------
    axis, zsum : 1D arrays
    """
    data_fmt = kws.pop("data_fmt", "curves")
    if data_fmt == "curves":
        ax, mat = curves_to_matrix(data, axis=axis, **kws)
    elif data_fmt == "lists":
        ax, mat = lists_to_matrix(data, axis=axis, **kws)
    else:
        raise NameError("'data_fmt' not understood")
    return ax, np.sum(mat, 0)


def avg_arrays_1d(data, axis=None, weights=None, **kws):
    """Average list of 1D arrays or curves by interpolation on a reference axis

    Parameters
    ----------
    data : lists of lists
    data_fmt : str
        define data format
        - "curves" -> :func:`curves_to_matrix`
        - "lists" -> :func:`lists_to_matrix`
    weights : None or array
        weights for the average

    Returns
    -------
    axis, zavg : 1D arrays
        np.average(zdats)
    """
    data_fmt = kws.pop("data_fmt", "curves")
    if data_fmt == "curves":
        ax, mat = curves_to_matrix(data, axis=axis, **kws)
    elif data_fmt == "lists":
        ax, mat = lists_to_matrix(data, axis=axis, **kws)
    else:
        raise NameError("'data_fmt' not understood")
    return ax, np.average(mat, axis=0, weights=weights)


def merge_arrays_1d(data, method="average", axis=None, weights=None, **kws):
    """Merge a list of 1D arrays by interpolation on a reference axis

    Parameters
    ----------
    data : lists of lists
    data_fmt : str
        define data format
        - "curves" -> :func:`curves_to_matrix`
        - "lists" -> :func:`curves_to_matrix`
    axis : None or array 1D, optional
        a reference axis used for the interpolation [None -> xdats[0]]
    method : str, optional
        method used to merge, available methods are:
            - "average" : uses np.average()
            - "sum" : uses np.sum()
    weights : None or array 1D, optional
        used if method == "average"

    Returns
    -------
    axis, zmrg : 1D arrays
        merge(zdats)
    """
    if method == "sum":
        return sum_arrays_1d(data, axis=axis, **kws)
    elif method == "average":
        return avg_arrays_1d(data, axis=axis, weights=weights, **kws)
    else:
        raise NameError("wrong 'method': %s" % method)


def rebin_piecewise_constant(x1, y1, x2):
    """Rebin histogram values y1 from old bin edges x1 to new edges x2.

    Code taken from: https://github.com/jhykes/rebin/blob/master/rebin.py

    It follows the procedure described in Figure 18.13 (chapter 18.IV.B.
    Spectrum Alignment, page 703) of Knoll [1]

    References
    ----------
    [1] Glenn Knoll, Radiation Detection and Measurement, third edition,
        Wiley, 2000.

    Parameters
    ----------
     - x1 : m+1 array of old bin edges.
     - y1 : m array of old histogram values.
            This is the total number in each bin, not an average.
     - x2 : n+1 array of new bin edges.

    Returns
    -------
     - y2 : n array of rebinned histogram values.
    """
    x1 = np.asarray(x1)
    y1 = np.asarray(y1)
    x2 = np.asarray(x2)

    # the fractional bin locations of the new bins in the old bins
    i_place = np.interp(x2, x1, np.arange(len(x1)))

    cum_sum = np.r_[[0], np.cumsum(y1)]

    # calculate bins where lower and upper bin edges span
    # greater than or equal to one original bin.
    # This is the contribution from the 'intact' bins (not including the
    # fractional start and end parts.
    whole_bins = np.floor(i_place[1:]) - np.ceil(i_place[:-1]) >= 1.0
    start = cum_sum[np.ceil(i_place[:-1]).astype(int)]
    finish = cum_sum[np.floor(i_place[1:]).astype(int)]

    y2 = np.where(whole_bins, finish - start, 0.0)

    bin_loc = np.clip(np.floor(i_place).astype(int), 0, len(y1) - 1)

    # fractional contribution for bins where the new bin edges are in the same
    # original bin.
    same_cell = np.floor(i_place[1:]) == np.floor(i_place[:-1])
    frac = i_place[1:] - i_place[:-1]
    contrib = frac * y1[bin_loc[:-1]]
    y2 += np.where(same_cell, contrib, 0.0)

    # fractional contribution for bins where the left and right bin edges are in
    # different original bins.
    different_cell = np.floor(i_place[1:]) > np.floor(i_place[:-1])
    frac_left = np.ceil(i_place[:-1]) - i_place[:-1]
    contrib = frac_left * y1[bin_loc[:-1]]

    frac_right = i_place[1:] - np.floor(i_place[1:])
    contrib += frac_right * y1[bin_loc[1:]]

    y2 += np.where(different_cell, contrib, 0.0)

    return y2


def reject_outliers(data, m=5.189, return_ma=False):
    """Reject outliers

    Modified from: https://stackoverflow.com/questions/11686720/is-there-a-numpy-builtin-to-reject-outliers-from-a-list
    See also: https://www.itl.nist.gov/div898/handbook/eda/section3/eda35h.htm
    """
    if not isinstance(data, np.ndarray):
        data = np.array(data)
    d = np.abs(data - np.median(data))
    mdev = np.median(d)
    s = d / (mdev if mdev else 1.0)
    mask = s < m
    if return_ma:
        imask = s > m
        return np.ma.masked_array(data=data, mask=imask)
    else:
        return data[mask]



#!/usr/bin/env python

"""
  Larch: a scientific data processing macro language based on python
"""
import os
import sys
import logging
import warnings
warnings.simplefilter('ignore')

logger = logging.getLogger()
logger.level = logging.WARNING

# note: may need to set CONDA env *before* loading numpy!
if os.name == 'nt':
    os.environ['CONDA_DLL_SEARCH_MODIFICATION_ENABLE'] = '1'

if (sys.version_info.major < 3 or sys.version_info.minor < 8):
    raise EnvironmentError('larch requires python 3.8 or higher')

import numpy
import scipy, scipy.optimize, scipy.special, scipy.stats
import matplotlib
import lmfit

# note: for HDF5 File / Filter Plugins to be useful, the
# hdf5plugin module needs to be imported before h5py
try:
    import hdf5plugin
    import h5py
except ImportError:
    pass

## be careful here: it is easy to have cicrular imports!

from .version import __date__, __version__, __release_version__
from .symboltable import Group, isgroup, repr_value
from .larchlib import Make_CallArgs, parse_group_args, isNamedClass, Journal, Entry
from .fitting import Parameter, isParameter, param_value, ParameterGroup

# from . import builtins
from .inputText import InputText
from .interpreter import Interpreter
from . import larchlib
from . import utils
from . import site_config



from typing import Optional
from . import sources
from . import hdf5_utils
from larch import Group


def open_xas_source(filename, **kw):
    with open(filename, "rb") as fh:
        topbytes = fh.read(10)

    if topbytes.startswith(b"\x89HDF\r"):
        creator = hdf5_utils.nexus_creator(filename).lower()
        class_name = None
        if creator == "bliss":
            class_name = "esrf"
        if not class_name:
            source = hdf5_utils.nexus_source(filename).lower()
            if "soleil" in source:
                class_name = "soleil"
        if not class_name:
            instrument = hdf5_utils.nexus_instrument(filename).lower()
            if "soleil" in instrument:
                class_name = "soleil"
        if not class_name:
            class_name = "nexus"
    elif topbytes.startswith(b"#S ") or topbytes.startswith(b"#F "):
        class_name = "spec"
    else:
        raise ValueError(f"Unknown file format: {filename}")
    return sources.get_source_type(class_name)(filename, **kw)


def read_xas_source(filename: str, scan: Optional[str] = None) -> Optional[Group]:
    if scan is None:
        return None
    source = open_xas_source(filename)
    scan = source.get_scan(scan)

    lgroup = Group(
        __name__=f"{source.TYPE} file: {filename}, scan: {scan.name}",
        filename=filename,
        source_info=source.get_source_info(),
        datatype="xas",
    )
    for name, value in scan._asdict().items():
        setattr(lgroup, name, value)
    for name, value in zip(scan.labels, scan.data):
        setattr(lgroup, name, value)
    return lgroup



#!/usr/bin/env python
"""
XANES Data Viewer and Analysis Tool
"""
import os
import sys
import time
import copy
import platform
from pathlib import Path
from importlib import import_module
from threading import Thread
import numpy as np
np.seterr(all='ignore')

from functools import partial

import wx
import wx.lib.scrolledpanel as scrolled
import wx.lib.agw.flatnotebook as flat_nb
from wx.adv import AboutBox, AboutDialogInfo

from wx.richtext import RichTextCtrl

import larch
from larch import Group, Journal, Entry
from larch.io import save_session, read_session
from larch.math import index_of
from larch.utils import (isotime, time_ago, get_cwd,
                         is_gzip, uname, path_split)
from larch.utils.strutils import (file2groupname, unique_name,
                                  common_startstring, asfloat, fix_varname)

from larch.larchlib import read_workdir, save_workdir, read_config, save_config

from larch.wxlib import (LarchFrame, ColumnDataFileFrame, AthenaImporter,
                         SpecfileImporter, XasImporter, FileCheckList, FloatCtrl,
                         FloatSpin, SetTip, get_icon, SimpleText, TextCtrl,
                         pack, Button, Popup, HLine, FileSave, FileOpen,
                         Choice, Check, MenuItem, HyperText, set_color, COLORS,
                         CEN, LEFT, FRAMESTYLE, Font, FONTSIZE,
                         flatnotebook, LarchUpdaterDialog, GridPanel,
                         CIFFrame, Structure2FeffFrame, FeffResultsFrame, LarchWxApp, OkCancel,
                         ExceptionPopup, set_color)


from larch.wxlib.plotter import get_display

from larch.fitting import fit_report
from larch.site_config import icondir, home_dir, user_larchdir
from larch.version import check_larchversion

from .xas_controller import XASController
from .taskpanel import GroupJournalFrame
from .config import (FULLCONF, CONF_SECTIONS,  CVar, ATHENA_CLAMPNAMES,
                     LARIX_PANELS, LARIX_MODES)

from .xas_dialogs import (MergeDialog, RenameDialog, RemoveDialog,
                          DeglitchDialog, ExportCSVDialog, RebinDataDialog,
                          EnergyCalibrateDialog, SmoothDataDialog,
                          OverAbsorptionDialog, DeconvolutionDialog,
                          SpectraCalcDialog,  QuitDialog, LoadSessionDialog,
                          fit_dialog_window)

from larch.io import (read_ascii, read_xdi, read_gsexdi, gsescan_group,
                          groups2csv, is_athena_project,
                      is_larch_session_file,
                      AthenaProject, make_hashkey, is_specfile, open_specfile)
from larch.io.xas_data_source import open_xas_source

from larch.xafs import pre_edge, pre_edge_baseline

# FNB_STYLE = flat_nb.FNB_NO_X_BUTTON
FNB_STYLE = flat_nb.FNB_X_ON_TAB
FNB_STYLE |= flat_nb.FNB_SMART_TABS|flat_nb.FNB_NO_NAV_BUTTONS

LEFT = wx.ALIGN_LEFT
CEN |=  wx.ALL
FILE_WILDCARDS = "Data Files|*.0*;*.dat;*.DAT;*.xdi;*.prj;*.sp*c;*.h*5;*.larix|All files (*.*)|*.*"

ICON_FILE = 'onecone.ico'
LARIX_SIZE = (1050, 850)
LARIX_MINSIZE = (500, 250)
PLOTWIN_SIZE = (550, 550)

QUIT_MESSAGE = '''Really Quit? You may want to save your project before quitting.
 This is not done automatically!'''

LARIX_TITLE = "Larix: XAS Visualization and Analysis"

def assign_gsescan_groups(group):
    labels = group.array_labels
    labels = []
    for i, name in enumerate(group.pos_desc):
        name = fix_varname(name.lower())
        labels.append(name)
        setattr(group, name, group.pos[i, :])

    for i, name in enumerate(group.sums_names):
        name = fix_varname(name.lower())
        labels.append(name)
        setattr(group, name, group.sums_corr[i, :])

    for i, name in enumerate(group.det_desc):
        name = fix_varname(name.lower())
        labels.append(name)
        setattr(group, name, group.det_corr[i, :])

    group.array_labels = labels


class PreferencesFrame(wx.Frame):
    """ edit preferences"""
    def __init__(self, parent, controller, **kws):
        self.controller = controller
        wx.Frame.__init__(self, None, -1,  'Larix Preferences',
                          style=FRAMESTYLE, size=(700, 725))

        sizer = wx.BoxSizer(wx.VERTICAL)
        tpanel = wx.Panel(self)

        self.title = SimpleText(tpanel, 'Edit Preference and Defaults',
                                size=(500, 25),
                                font=Font(FONTSIZE+1), style=LEFT,
                                colour=COLORS['nb_text'])

        self.save_btn = Button(tpanel, 'Save for Future sessions',
                               size=(200, -1), action=self.onSave)

        self.nb = flatnotebook(tpanel, {})
        self.wids = {}
        conf = self.controller.config

        def text(panel, label, size):
            return SimpleText(panel, label, size=(size, -1), style=LEFT)

        for name, data in FULLCONF.items():
            self.wids[name] = {}

            panel = GridPanel(self.nb, ncols=3, nrows=8, pad=3, itemstyle=LEFT)
            panel.SetFont(Font(FONTSIZE))
            title = CONF_SECTIONS.get(name, name)
            title = SimpleText(panel, f"  {title}",
                               size=(550, -1), font=Font(FONTSIZE+2),
                               colour=COLORS['title'], style=LEFT)

            self.wids[name]['_key_'] = SimpleText(panel, " <name> ",
                                                  size=(150, -1), style=LEFT)
            self.wids[name]['_help_'] = SimpleText(panel, " <click on name for description> ",
                                                   size=(525, 30), style=LEFT)

            panel.Add((5, 5), newrow=True)
            panel.Add(title, dcol=4)
            panel.Add((5, 5), newrow=True)
            panel.Add(self.wids[name]['_key_'])
            panel.Add(self.wids[name]['_help_'],  dcol=3)
            panel.Add((5, 5), newrow=True)
            panel.Add(HLine(panel, (625, 3)), dcol=4)

            panel.Add((5, 5), newrow=True)
            panel.Add(text(panel, 'Name', 150))

            panel.Add(text(panel, 'Value', 250))
            panel.Add(text(panel, 'Factory Default Value', 225))

            for key, cvar in data.items():
                val = conf[name][key]
                cb = partial(self.update_value, section=name, option=key)
                helpcb = partial(self.update_help, section=name, option=key)
                wid = None
                if cvar.dtype == 'choice':
                    wid = Choice(panel, size=(250, -1), choices=cvar.choices, action=cb)
                    if not isinstance(val, str): val = str(val)
                    wid.SetStringSelection(val)
                elif cvar.dtype == 'bool':
                    wid = Choice(panel, size=(250, -1), choices=['True', 'False'], action=cb)
                    wid.SetStringSelection('True' if val else 'False')
                elif cvar.dtype in ('int', 'float'):
                    digits = 2 if cvar.dtype == 'float' else 0
                    wid = FloatSpin(panel, value=val, min_val=cvar.min, max_val=cvar.max,
                                  digits=digits, increment=cvar.step, size=(250, -1), action=cb)
                else:
                    wid = TextCtrl(panel, size=(250, -1), value=val, action=cb)

                label = HyperText(panel, key, action=helpcb, size=(150, -1))
                panel.Add((5, 5), newrow=True)
                panel.Add(label)
                panel.Add(wid)
                panel.Add(text(panel, f'{cvar.value}', 225))
                SetTip(wid, cvar.desc)
                self.wids[name][key] = wid

            panel.pack()
            self.nb.AddPage(panel, name, True)

        self.nb.SetSelection(0)

        sizer.Add(self.title, 0, LEFT, 3)
        sizer.Add(self.save_btn, 0, LEFT, 5)
        sizer.Add((5, 5), 0, LEFT, 5)
        sizer.Add(self.nb, 1, LEFT|wx.EXPAND, 5)
        pack(tpanel, sizer)
        w0, h0 = self.GetSize()
        w1, h1 = self.GetBestSize()
        self.SetSize((max(w0, w1)+25, max(h0, h1)+25))

        self.Show()
        self.Raise()

    def update_help(self, label=None, event=None, section='main', option=None):
        cvar = FULLCONF[section][option]
        self.wids[section]['_key_'].SetLabel("%s : " % option)
        self.wids[section]['_help_'].SetLabel(cvar.desc)

    def update_value(self, event=None, section='main', option=None):
        cvar = FULLCONF[section][option]
        wid = self.wids[section][option]
        value = cvar.value
        if cvar.dtype == 'bool':
            value = wid.GetStringSelection().lower().startswith('t')
        elif cvar.dtype == 'choice':
            value = wid.GetStringSelection()
        elif cvar.dtype == 'int':
            value = int(wid.GetValue())
        elif cvar.dtype == 'float':
            value = float(wid.GetValue())
        else:
            value = wid.GetValue()
        self.controller.config[section][option] = value

    def onSave(self, event=None):
        self.controller.save_config()

class PanelSelectionFrame(wx.Frame):
    """select analysis panels to display"""
    def __init__(self, parent, controller, **kws):
        self.controller = controller
        self.parent = parent
        wx.Frame.__init__(self, None, -1,  'Larix Configuration: Analysis Panels',
                          style=FRAMESTYLE, size=(650, 500))
        self.wids = {}

        style    = wx.DEFAULT_FRAME_STYLE|wx.TAB_TRAVERSAL
        labstyle  = wx.ALIGN_LEFT|wx.ALIGN_CENTER_VERTICAL|wx.ALL
        rlabstyle = wx.ALIGN_RIGHT|wx.ALIGN_CENTER_VERTICAL|wx.ALL
        tstyle    = wx.ALIGN_LEFT|wx.ALIGN_CENTER_VERTICAL

        font = parent.GetFont()

        titlefont  = self.GetFont()
        titlefont.PointSize += 2
        titlefont.SetWeight(wx.BOLD)

        sizer = wx.GridBagSizer(5, 5)
        panel = scrolled.ScrolledPanel(self, size=(700, 750),
                                       style=wx.GROW|wx.TAB_TRAVERSAL)

        title = SimpleText(panel, 'Select Larix Modes and Analysis Panels',
                           size=(500, 25),
                           font=Font(FONTSIZE+1), style=LEFT,
                           colour=COLORS['nb_text'])
        modetitle = SimpleText(panel, 'Analysis Mode: ')
        panetitle = SimpleText(panel, 'Select Individual Analysis Panels: ')

        self.wids = wids = {}
        self.current_mode = self.parent.mode
        modenames = [m[0] for m in LARIX_MODES.values()]

        wids['modechoice'] = Choice(panel, choices=modenames, size=(350, -1),
                                    action=self.on_modechoice)
        wids['modechoice'].SetStringSelection(LARIX_MODES[self.current_mode][0])

        page_map = self.parent.get_panels()
        irow = 0
        sizer.Add(title, (irow, 0), (1, 2), labstyle|wx.ALL, 3)
        irow = 1
        sizer.Add(modetitle, (irow, 0), (1, 1), labstyle|wx.ALL, 3)
        sizer.Add(wids['modechoice'], (irow, 1), (1, 1), labstyle|wx.ALL, 3)
        irow += 1
        sizer.Add(HLine(panel, (325, 3)), (irow, 0), (1, 2), labstyle|wx.ALL, 3)
        irow += 1
        sizer.Add(panetitle, (irow, 0), (1, 2), labstyle|wx.ALL, 3)

        self.selections = {}
        strlen = 30
        for pagename in page_map:
            strlen = max(strlen, len(pagename))

        for key, atab in LARIX_PANELS.items():
            iname = (atab.title + ' '*strlen)[:strlen]
            cb = wx.CheckBox(panel, -1, iname)#, style=wx.ALIGN_RIGHT)
            cb.SetValue(atab.title in page_map)
            desc = SimpleText(panel, LARIX_PANELS[key].desc)
            self.selections[key] = cb
            irow += 1
            sizer.Add(cb, (irow, 0), (1, 1), labstyle,  5)
            sizer.Add(desc, (irow, 1), (1, 1), labstyle,  5)

        irow += 1
        sizer.Add(HLine(panel, (325, 3)), (irow, 0), (1, 2), labstyle|wx.ALL, 3)

        btn_ok     = Button(panel, 'Apply', size=(70, -1), action=self.OnOK)
        btn_cancel = Button(panel, 'Done', size=(70, -1), action=self.OnCancel)

        irow += 1
        sizer.Add(btn_ok,     (irow, 0), (1, 1), labstyle|wx.ALL, 3)
        sizer.Add(btn_cancel, (irow, 1), (1, 1), labstyle|wx.ALL, 3)

        pack(panel, sizer)
        panel.SetupScrolling()
        mainsizer = wx.BoxSizer(wx.VERTICAL)
        mainsizer.Add(panel, 1, wx.GROW|wx.ALL, 1)

        self.SetMinSize((750, 450))
        pack(self, mainsizer)
        self.Show()
        self.Raise()

    def on_modechoice(self, event=None):
        modename = event.GetString()
        self.current_mode = 'xas'
        for key, dat in LARIX_MODES.items():
            if dat[0] == modename:
                self.current_mode = key
        panels = LARIX_MODES[self.current_mode][1]
        self.Freeze()
        for name in self.selections:
            self.selections[name].SetValue(False)

        for name in panels:
            if name in self.selections:
                self.selections[name].SetValue(True)
        self.Thaw()

    def OnOK(self, event=None):
        self.parent.Hide()
        self.parent.Freeze()
        for i in range(self.parent.nb.GetPageCount()):
            self.parent.nb.DeletePage(0)

        for name, cb in self.selections.items():
            if cb.IsChecked():
                self.parent.add_analysis_panel(name)
        self.parent.nb.SetSelection(0)
        self.parent.mode = self.current_mode
        self.parent.Thaw()
        self.parent.Show()

    def OnCancel(self, event=None):
        self.Destroy()

class LarixFrame(wx.Frame):
    _about = f"""{LARIX_TITLE}
    Matt Newville <newville @ cars.uchicago.edu>
    """
    def __init__(self, parent=None, _larch=None, filename=None,
                 with_wx_inspect=False, mode='xas', check_version=True, **kws):
        wx.Frame.__init__(self, parent, -1, size=LARIX_SIZE, style=FRAMESTYLE)

        if check_version:
            def version_checker():
                self.vinfo = check_larchversion()
            version_thread = Thread(target=version_checker)
            version_thread.start()

        self.with_wx_inspect = with_wx_inspect
        self.mode = mode
        if self.mode not in LARIX_MODES:
            self.mode = 'xas'
        self.last_col_config = {}
        self.last_spec_config = {}
        self.last_session_file = None
        self.last_session_read = None
        self.last_athena_file = None
        self.paths2read = []
        self.current_filename = filename
        title = LARIX_TITLE

        self.larch_buffer = parent
        if not isinstance(parent, LarchFrame):
            self.larch_buffer = LarchFrame(_larch=_larch,
                                           parent=self,
                                           is_standalone=False,
                                           with_raise=False,
                                           exit_on_close=False)

        self.larch = self.larch_buffer.larchshell

        self.controller = XASController(wxparent=self, _larch=self.larch)
        iconfile = Path(icondir, ICON_FILE).as_posix()
        self.SetIcon(wx.Icon(iconfile, wx.BITMAP_TYPE_ICO))

        self.last_autosave = 0
        self.last_save_message = ('Session has not been saved', '', '')


        self.timers = {'pin': wx.Timer(self),
                       'autosave': wx.Timer(self)}
        self.Bind(wx.EVT_TIMER, self.onPinTimer, self.timers['pin'])
        self.Bind(wx.EVT_TIMER, self.onAutoSaveTimer, self.timers['autosave'])
        self.cursor_dat = {}

        self.subframes = {}
        self.plotframe = None
        self.SetTitle(title)
        self.SetSize(LARIX_SIZE)
        self.SetMinSize(LARIX_MINSIZE)
        self.SetFont(Font(FONTSIZE))
        self.createMainPanel()
        self.createMenus()
        self.statusbar = self.CreateStatusBar(2, style=wx.STB_DEFAULT_STYLE)
        self.statusbar.SetStatusWidths([-3, -1])
        statusbar_fields = [" ", "ready"]
        for i in range(len(statusbar_fields)):
            self.statusbar.SetStatusText(statusbar_fields[i], i)
        self.Show()

        self.Raise()
        self.statusbar.SetStatusText('ready', 1)
        self.timers['autosave'].Start(30_000)

        if self.current_filename is not None:
            wx.CallAfter(self.onRead, self.current_filename)

        # show_wxsizes(self)
        if check_version:
            version_thread.join()
            if self.vinfo is not None:
                if self.vinfo.update_available:
                    self.statusbar.SetStatusText(f'Larch Version {self.vinfo.remote_version} is available!', 0)
                    self.statusbar.SetStatusText(f'Larch Version {self.vinfo.local_version}', 1)
                else:
                    self.statusbar.SetStatusText(f'Larch Version {self.vinfo.local_version} (latest)', 1)
        xpos, ypos = self.GetPosition()
        xsiz, ysiz = self.GetSize()
        plotpos = (xpos+xsiz+2, ypos)
        self.controller.get_display(stacked=False, position=plotpos)
        self.Raise()


    def createMainPanel(self):
        display0 = wx.Display(0)
        client_area = display0.ClientArea
        xmin, ymin, xmax, ymax = client_area
        xpos = int((xmax-xmin)*0.02) + xmin
        ypos = int((ymax-ymin)*0.04) + ymin
        self.SetPosition((xpos, ypos))

        splitter  = wx.SplitterWindow(self, style=wx.SP_LIVE_UPDATE,
                                      size=(700, 700))
        splitter.SetMinimumPaneSize(250)

        leftpanel = wx.Panel(splitter)
        ltop = wx.Panel(leftpanel)

        def Btn(msg, x, act):
            b = Button(ltop, msg, size=(x, 30),  action=act)
            b.SetFont(Font(FONTSIZE))
            return b

        sel_none = Btn('Select None',   120, self.onSelNone)
        sel_all  = Btn('Select All',    120, self.onSelAll)

        file_actions = [('Show Group Journal', self.onGroupJournal),
                        ('Copy Group', self.onCopyGroup),
                        ('Rename Group', self.onRenameGroup),
                        ('Remove Group', self.onRemoveGroup)]

        self.controller.filelist = FileCheckList(leftpanel, main=self,
                                                 pre_actions=file_actions,
                                                 select_action=self.ShowFile,
                                                 remove_action=self.RemoveFile)
        set_color(self.controller.filelist, 'list_fg', bg='list_bg')
        # self.controller.filelist.check_event  = self.filelist_check_event
        self.controller.filelist.Bind(wx.EVT_CHECKLISTBOX, self.filelist_check_event)

        tsizer = wx.BoxSizer(wx.HORIZONTAL)
        tsizer.Add(sel_all, 1, LEFT|wx.GROW, 1)
        tsizer.Add(sel_none, 1, LEFT|wx.GROW, 1)
        pack(ltop, tsizer)

        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add(ltop, 0, LEFT|wx.GROW, 1)
        sizer.Add(self.controller.filelist, 1, LEFT|wx.GROW|wx.ALL, 1)

        pack(leftpanel, sizer)

        # right hand side
        panel = scrolled.ScrolledPanel(splitter)
        panel.SetSize((650, 650))
        panel.SetMinSize((450, 550))
        sizer = wx.BoxSizer(wx.VERTICAL)
        self.title = SimpleText(panel, ' ', size=(500, 25),
                                font=Font(FONTSIZE+3), style=LEFT,
                                colour=COLORS['nb_text'])

        ir = 0
        sizer.Add(self.title, 0, CEN, 3)

        self.nb = flatnotebook(panel, {},
                               panelkws={'controller': self.controller},
                               on_change=self.onNBChanged,
                               style=FNB_STYLE,
                               size=(700, 700))
        panels = LARIX_MODES[self.mode][1]
        for key, atab in LARIX_PANELS.items():
            if key in panels:
                self.add_analysis_panel(key)
        self.nb.SetSelection(0)

        sizer.Add(self.nb, 1, LEFT|wx.EXPAND, 2)
        panel.SetupScrolling()

        pack(panel, sizer)
        splitter.SplitVertically(leftpanel, panel, 1)

    def get_panels(self):
        "return current mapping of displayed panels"
        out = {}
        for i in range(self.nb.GetPageCount()):
            out[self.nb.GetPageText(i)] = i
        return out

    def add_analysis_panel(self, name):
        """make sure an analysis panel is displayed ,
           and return its current index in the set of panels
           or None to signal "unknonwn panel name"
        """
        atab = LARIX_PANELS.get(name, None)
        if atab is None:
            return None

        current_panels = self.get_panels()
        #print("Add panel , current = ", name, atab.title)
        if atab.title in current_panels:
            return current_panels[atab.title]
        # not in current tabs, so add it
        cons = atab.constructor.split('.')
        clsname = cons.pop()
        module = '.'.join(cons)
        try:
            cls = getattr(import_module(module), clsname)
        except (ImportError, KeyError):
            cls = None
            print(f"cannot find analysis panel {atab}")
        if cls is not None:
            # print("Add panel , current = ", name, atab)
            nbpanel = cls(parent=self, controller=self.controller)
            self.nb.AddPage(nbpanel, atab.title, True)
        current_panels = self.get_panels()
        return current_panels.get(atab.title, None)

    def process_normalization(self, dgroup, force=True, use_form=True, force_mback=False):
        self.get_nbpage('xasnorm')[1].process(dgroup, force=force,
                                              force_mback=False, use_form=use_form)

    def process_exafs(self, dgroup, force=True):
        self.get_nbpage('exafs')[1].process(dgroup, force=force)

    def get_nbpage(self, name):
        "get nb page by name"
        name = name.lower()
        out = 0
        if name not in LARIX_PANELS:
            print("unknown panel : ", name, LARIX_PANELS)
            return 0, self.nb.GetPage(0)
        # print("GET NB PAGE ", name, LARIX_PANELS.get(name, 'gg'))

        atab = LARIX_PANELS[name]
        current_panels = self.get_panels()
        if atab.title in current_panels:
            ipage = current_panels[atab.title]
        else:
            ipage = self.add_analysis_panel(name)
        # print("GET NB PAGE ", ipage, self.nb.GetPage(ipage))
        return ipage, self.nb.GetPage(ipage)

    def onNBChanged(self, event=None):
        callback = getattr(self.nb.GetCurrentPage(), 'onPanelExposed', None)
        if callable(callback):
            callback()

    def onSelAll(self, event=None):
        self.controller.filelist.select_all()

    def onSelNone(self, event=None):
        self.controller.filelist.select_none()

    def write_message(self, msg, panel=0):
        """write a message to the Status Bar"""
        self.statusbar.SetStatusText(msg, panel)

    def RemoveFile(self, fname=None, **kws):
        if fname is not None:
            s = str(fname)
            if s in self.controller.file_groups:
                group = self.controller.file_groups.pop(s)
            self.controller.sync_xasgroups()

    def filelist_check_event(self, evt=None):
        """MN 2024-Feb this is included to better 'swallow' the checked event,
        so that it does in fact run ShowFile().
        This could be removed eventually, as wxutils will also no longer run
        filelist.SetSelection()"""
        index = evt.GetSelection()
        label = evt.GetString()
        pass

    def ShowFile(self, evt=None, groupname=None, process=True,
                 filename=None, plot=True, **kws):
        if filename is None and evt is not None:
            filename = str(evt.GetString())

        if groupname is None and filename is not None:
            groupname = self.controller.file_groups[filename]

        if not hasattr(self.larch.symtable, groupname):
            return

        dgroup = self.controller.get_group(groupname)
        if dgroup is None:
            return

        # print("This ShowFile ", groupname, getattr(dgroup, 'datatype', 'xydata'))

        if (getattr(dgroup, 'datatype', 'xydata').startswith('xa') and not
            (hasattr(dgroup, 'norm') and hasattr(dgroup, 'e0'))):
            self.process_normalization(dgroup, force=True, use_form=False)
        if filename is None:
            filename = dgroup.filename
        self.current_filename = filename
        journal = getattr(dgroup, 'journal', Journal(source_desc=filename))
        if isinstance(journal, Journal):
            sdesc = journal.get('source_desc', latest=True)
        else:
            sdesc = journal.get('source_desc', '?')

        if isinstance(sdesc, Entry):
            sdesc = sdesc.value
        if not isinstance(sdesc, str):
            sdesc = repr(sdesc)
        self.title.SetLabel(sdesc)

        self.controller.group = dgroup
        self.controller.groupname = groupname
        cur_panel = self.nb.GetCurrentPage()
        # print("Got CUR PANEL  ", cur_panel)
        if process:
            cur_panel.fill_form(dgroup)
            cur_panel.skip_process = False
            cur_panel.process(dgroup=dgroup)
            if plot and hasattr(cur_panel, 'plot'):
                cur_panel.plot(dgroup=dgroup)
            cur_panel.skip_process = False

        self.controller.filelist.SetStringSelection(filename)


    def createMenus(self):
        # ppnl = self.plotpanel
        self.menubar = wx.MenuBar()
        file_menu = wx.Menu()
        session_menu = wx.Menu()
        pref_menu = wx.Menu()
        group_menu = wx.Menu()
        xasdata_menu = wx.Menu()
        feff_menu = wx.Menu()
        m = {}

        MenuItem(self, file_menu, "&Open Data File\tCtrl+O",
                 "Open Data File",  self.onReadDialog)

        file_menu.AppendSeparator()

        MenuItem(self, file_menu, "&Read Larch Session\tCtrl+R",
                 "Read Previously Saved Session",  self.onLoadSession)

        MenuItem(self, file_menu, "&Save Larch Session\tCtrl+S",
                 "Save Session to a File",  self.onSaveSession)

        MenuItem(self, file_menu, "&Save Larch Session As ...\tCtrl+A",
                 "Save Session to a File",  self.onSaveSessionAs)


        file_menu.AppendSeparator()
        MenuItem(self, file_menu, "Save Selected Groups to Athena Project File",
                 "Save Selected Groups to an Athena Project File",
                 self.onExportAthenaProject)

        MenuItem(self, file_menu, "Save Selected Groups to CSV File",
                 "Save Selected Groups to a CSV File",
                 self.onExportCSV)

        MenuItem(self, file_menu, "&Quit\tCtrl+Q", "Quit program", self.onClose)


        MenuItem(self, session_menu, "&Read Larch Session",
                 "Read Previously Saved Session",  self.onLoadSession)

        MenuItem(self, session_menu, "&Save Larch Session",
                 "Save Session to a File",  self.onSaveSession)

        MenuItem(self, session_menu, "&Save Larch Session As ...",
                 "Save Session to a File",  self.onSaveSessionAs)

        MenuItem(self, session_menu, "Clear Larch Session",
                 "Clear all data from this Session",  self.onClearSession)

        # autosaved session
        conf = self.controller.get_config('autosave',
                                          {'fileroot': 'autosave'})
        froot= conf['fileroot']

        self.recent_menu = wx.Menu()
        self.get_recent_session_menu()
        session_menu.Append(-1, 'Recent Session Files',  self.recent_menu)


        MenuItem(self, session_menu, "&Auto-Save Larch Session",
                 f"Save Session now",  self.autosave_session)

        session_menu.AppendSeparator()

        MenuItem(self, session_menu, 'Show Larch Buffer\tCtrl+L',
                 'Show Larch Programming Buffer',
                 self.onShowLarchBuffer)

        MenuItem(self, session_menu, 'Save Larch History as Script\tCtrl+H',
                 'Save Session History as Larch Script',
                 self.onSaveLarchHistory)

        if self.with_wx_inspect:
            MenuItem(self, session_menu, 'wx inspect\tCtrl+I',
                     'Show wx inspection window',   self.onwxInspect)

        MenuItem(self, pref_menu, 'Select Analysis Panels and Modes',
                'Select Analysis Panels and Modes',
                 self.onPanelSelect)

        MenuItem(self, pref_menu, 'Edit Preferences\tCtrl+E', 'Customize Preferences',
                 self.onPreferences)

        MenuItem(self, group_menu, "Copy This Group",
                 "Copy This Group", self.onCopyGroup)

        MenuItem(self, group_menu, "Rename This Group",
                 "Rename This Group", self.onRenameGroup)

        MenuItem(self, group_menu, "Show Journal for This Group",
                 "Show Processing Journal for This Group", self.onGroupJournal)


        group_menu.AppendSeparator()

        MenuItem(self, group_menu, "Remove Selected Groups",
                 "Remove Selected Group", self.onRemoveGroups)

        group_menu.AppendSeparator()

        MenuItem(self, group_menu, "Merge Selected Groups",
                 "Merge Selected Groups", self.onMergeData)

        group_menu.AppendSeparator()

        MenuItem(self, group_menu, "Freeze Selected Groups",
                 "Freeze Selected Groups", self.onFreezeGroups)

        MenuItem(self, group_menu, "UnFreeze Selected Groups",
                 "UnFreeze Selected Groups", self.onUnFreezeGroups)

        MenuItem(self, xasdata_menu, "Deglitch Data",  "Deglitch Data",
                 self.onDeglitchData)

        MenuItem(self, xasdata_menu, "Calibrate Energy",
                 "Calibrate Energy",
                 self.onEnergyCalibrateData)

        MenuItem(self, xasdata_menu, "Smooth Data", "Smooth Data",
                 self.onSmoothData)

        MenuItem(self, xasdata_menu, "Deconvolve Data",
                 "Deconvolution of Data",  self.onDeconvolveData)

        MenuItem(self, xasdata_menu, "Rebin Data", "Rebin Data",
                 self.onRebinData)

        MenuItem(self, xasdata_menu, "Correct Over-absorption",
                 "Correct Over-absorption",
                 self.onCorrectOverAbsorptionData)

        MenuItem(self, xasdata_menu, "Add and Subtract Spectra",
                 "Calculations of Spectra",  self.onSpectraCalc)


        self.menubar.Append(file_menu, "&File")
        self.menubar.Append(session_menu, "Sessions")
        self.menubar.Append(pref_menu, "Preferences")
        self.menubar.Append(group_menu, "Groups")
        self.menubar.Append(xasdata_menu, "XAS Data")

        MenuItem(self, feff_menu, "Browse CIF Structures, Run Feff",
                 "Browse CIF Structure, run Feff", self.onCIFBrowse)
        MenuItem(self, feff_menu, "Generate Feff input from general structures, Run Feff",
                 "Generate Feff input from general structures, run Feff", self.onStructureBrowse)
        MenuItem(self, feff_menu, "Browse Feff Calculations",
                 "Browse Feff Calculations, Get Feff Paths", self.onFeffBrowse)

        self.menubar.Append(feff_menu, "Feff")

        hmenu = wx.Menu()
        MenuItem(self, hmenu, 'About Larix', 'About Larix',
                 self.onAbout)
        MenuItem(self, hmenu, 'Check for Updates', 'Check for Updates',
                 self.onCheckforUpdates)

        self.menubar.Append(hmenu, '&Help')
        self.SetMenuBar(self.menubar)
        self.Bind(wx.EVT_CLOSE,  self.onClose)

    def onwxInspect(self, evt=None):
        wx.GetApp().ShowInspectionTool()

    def onShowLarchBuffer(self, evt=None):
        if self.larch_buffer is None:
            self.larch_buffer = LarchFrame(_larch=self.larch, is_standalone=False)
        self.larch_buffer.Show()
        self.larch_buffer.Raise()

    def onSaveLarchHistory(self, evt=None):
        wildcard = 'Larch file (*.lar)|*.lar|All files (*.*)|*.*'
        path = FileSave(self, message='Save Session History as Larch Script',
                        wildcard=wildcard,
                        default_file='larix_history.lar')
        if path is not None:
            self.larch._larch.input.history.save(path, session_only=True)
            self.write_message("Wrote history %s" % path, 0)

    def onExportCSV(self, evt=None):
        filenames = self.controller.filelist.GetCheckedStrings()
        if len(filenames) < 1:
            Popup(self, "No files selected to export to CSV",
                  "No files selected")
            return

        deffile = "%s_%i.csv" % (filenames[0], len(filenames))

        dlg = ExportCSVDialog(self, filenames)
        res = dlg.GetResponse()

        dlg.Destroy()
        if not res.ok:
            return

        deffile = f"{filenames[0]:s}_{len(filenames):d}.csv"
        wcards  = 'CSV Files (*.csv)|*.csv|All files (*.*)|*.*'

        outfile = FileSave(self, 'Save Groups to CSV File',
                           default_file=deffile, wildcard=wcards)

        if outfile is None:
            return
        if Path(outfile).exists() and uname != 'darwin':  # darwin prompts in FileSave!
            if wx.ID_YES != Popup(self,
                                  "Overwrite existing CSV File?",
                                  "Overwrite existing file?", style=wx.YES_NO):
                return

        savegroups = [self.controller.filename2group(res.master)]
        for fname in filenames:
            dgroup = self.controller.filename2group(fname)
            if dgroup not in savegroups:
                savegroups.append(dgroup)

        try:
            groups2csv(savegroups, outfile, x=res.xarray, y=res.yarray,
                    delim=res.delim, individual=res.individual)
            self.write_message(f"Exported CSV file {outfile:s}")
        except:
            title = "Could not export CSV File"
            message = [f"Could not export CSV File {outfile}"]
            ExceptionPopup(self, title, message)

    # Athena
    def onExportAthenaProject(self, evt=None):
        groups = []
        self.controller.sync_xasgroups()
        for checked in self.controller.filelist.GetCheckedStrings():
            groups.append(self.controller.file_groups[str(checked)])

        if len(groups) < 1:
             Popup(self, "No files selected to export to Project",
                   "No files selected")
             return
        prompt, prjfile = self.get_athena_project()
        self.save_athena_project(prjfile, groups)

    def get_athena_project(self):
        prjfile = self.last_athena_file
        prompt = False
        if prjfile is None:
            tstamp = isotime(filename=True)[:15]
            prjfile = f"{tstamp:s}.prj"
            prompt = True
        return prompt, prjfile

    def onSaveAthenaProject(self, evt=None):
        groups = self.controller.filelist.GetItems()
        if len(groups) < 1:
            Popup(self, "No files to export to Project", "No files to export")
            return

        prompt, prjfile = self.get_athenaproject()
        self.save_athena_project(prjfile, groups, prompt=prompt,
                                 warn_overwrite=False)

    def onSaveAsAthenaProject(self, evt=None):
        groups = self.controller.filelist.GetItems()
        if len(groups) < 1:
            Popup(self, "No files to export to Project", "No files to export")
            return

        prompt, prjfile = self.get_athena_project()
        self.save_athena_project(prjfile, groups)

    def save_athena_project(self, filename, grouplist, prompt=True,
                            warn_overwrite=True):
        if len(grouplist) < 1:
            return
        savegroups = [self.controller.get_group(gname) for gname in grouplist]
        if prompt:
            filename = Path(filename).name
            wcards  = 'Project Files (*.prj)|*.prj|All files (*.*)|*.*'
            filename = FileSave(self, 'Save Groups to Project File',
                                default_file=filename, wildcard=wcards)
            if filename is None:
                return

        if (Path(filename).exists() and warn_overwrite and
            uname != 'darwin'):  # darwin prompts in FileSave!
            if wx.ID_YES != Popup(self,
                                  "Overwrite existing Project File?",
                                  "Overwrite existing file?", style=wx.YES_NO):
                return

        aprj = AthenaProject(filename=filename)
        for label, grp in zip(grouplist, savegroups):
            aprj.add_group(grp)
        aprj.save(use_gzip=True)
        self.write_message("Saved project file %s" % (filename))
        self.last_athena_file = filename


    def onPreferences(self, evt=None):
        self.show_subframe('preferences', PreferencesFrame,
                           controller=self.controller)

    def onPanelSelect(self, evt=None):
        self.show_subframe('panel_select', PanelSelectionFrame,
                           controller=self.controller)

    def onLoadSession(self, evt=None, path=None):
        if path is None:
            wildcard = 'Larch Session File (*.larix)|*.larix|All files (*.*)|*.*'
            path = FileOpen(self, message="Load Larch Session",
                            wildcard=wildcard, default_file='larch.larix')
        if path is None:
            return

        if is_athena_project(path):
            self.show_subframe('athena_import', AthenaImporter,
                               controller=self.controller, filename=path,
                               read_ok_cb=self.onReadAthenaProject_OK)
            return

        try:
            _session  = read_session(path)
        except:
            title = "Invalid Path for Larch Session"
            message = [f"{path} is not a valid Larch Session File"]
            ExceptionPopup(self, title, message)
            return

        LoadSessionDialog(self, _session, path, self.controller).Show()
        self.last_session_read = path
        fpath = Path(path).absolute()
        fname = fpath.name
        fdir = fpath.parent.as_posix()
        if self.controller.chdir_on_fileopen() and len(fdir) > 0:
            os.chdir(fdir)
            self.controller.set_workdir()

    def onSaveSessionAs(self, evt=None):
        groups = self.controller.filelist.GetItems()
        if len(groups) < 1:
            return
        self.last_session_file = None
        self.onSaveSession()


    def onSaveSession(self, evt=None):
        groups = self.controller.filelist.GetItems()
        if len(groups) < 1:
            return

        fname = self.last_session_file
        if fname is None:
            fname = self.last_session_read
            if fname is None:
                fname = time.strftime('%Y%b%d_%H%M') + '.larix'

            fname = Path(fname).name
            wcards  = 'Larch Project Files (*.larix)|*.larix|All files (*.*)|*.*'
            fname = FileSave(self, 'Save Larch Session File',
                             default_file=fname, wildcard=wcards)
            if fname is None:
                return

            if Path(fname).exists() and uname != 'darwin':  # darwin prompts in FileSave!
                if wx.ID_YES != Popup(self, "Overwrite existing Project File?",
                                      "Overwrite existing file?", style=wx.YES_NO):
                    return

        save_session(fname=fname, _larch=self.larch._larch)
        stime = time.strftime("%H:%M")
        self.last_save_message = ("Session last saved", f"'{fname}'", f"{stime}")
        self.write_message(f"Saved session to '{fname}' at {stime}")
        self.last_session_file = self.last_session_read = fname

    def onClearSession(self, evt=None):
        conf = self.controller.get_config('autosave',
                                          {'fileroot': 'autosave'})
        afile = Path(self.controller.larix_folder,
                         conf['fileroot']+'.larix').as_posix()

        msg = f"""Session will be saved to
         '{afile:s}'
before clearing"""

        dlg = wx.Dialog(None, -1, title="Clear all Session data?", size=(550, 300))
        dlg.SetFont(Font(FONTSIZE))
        panel = GridPanel(dlg, ncols=3, nrows=4, pad=2, itemstyle=LEFT)

        panel.Add(wx.StaticText(panel, label="Clear all Session Data?"), dcol=2)
        panel.Add(wx.StaticText(panel, label=msg), dcol=4, newrow=True)

        panel.Add((5, 5) ,  newrow=True)
        panel.Add((5, 5), newrow=True)
        panel.Add(OkCancel(panel), dcol=2, newrow=True)
        panel.pack()

        fit_dialog_window(dlg, panel)


        if wx.ID_OK == dlg.ShowModal():
            self.autosave_session()
            self.controller.clear_session()
        dlg.Destroy()


    def onConfigDataProcessing(self, event=None):
        pass


    def onCopyGroup(self, event=None, journal=None):
        fname = self.current_filename
        if fname is None:
            fname = self.controller.filelist.GetStringSelection()
        ogroup = self.controller.get_group(fname)
        ngroup = self.controller.copy_group(fname)
        self.install_group(ngroup, journal=ogroup.journal)

    def onGroupJournal(self, event=None):
        dgroup = self.controller.get_group()
        if dgroup is not None:
            self.show_subframe('group_journal', GroupJournalFrame)
            self.subframes['group_journal'].set_group(dgroup)


    def onRenameGroup(self, event=None):
        fname = self.current_filename = self.controller.filelist.GetStringSelection()
        if fname is None:
            return
        dlg = RenameDialog(self, fname)
        res = dlg.GetResponse()
        dlg.Destroy()

        if res.ok:
            selected = []
            for checked in self.controller.filelist.GetCheckedStrings():
                selected.append(str(checked))
            if self.current_filename in selected:
                selected.remove(self.current_filename)
                selected.append(res.newname)

            groupname = self.controller.file_groups.pop(fname)
            self.controller.sync_xasgroups()
            self.controller.file_groups[res.newname] = groupname
            self.controller.filelist.rename_item(self.current_filename, res.newname)
            dgroup = self.controller.get_group(groupname)
            dgroup.filename = self.current_filename = res.newname

            self.controller.filelist.SetCheckedStrings(selected)
            self.controller.filelist.SetStringSelection(res.newname)

    def onRemoveGroup(self, event=None):
        n = int(self.controller.filelist.GetSelection())
        all_names = self.controller.filelist.GetItems()
        fname = all_names[n]

        do_remove = (wx.ID_YES == Popup(self,
                                        f"Remove Group '{fname}'?",
                                        'Remove Group? Cannot be undone!',
                                        style=wx.YES_NO))
        if do_remove:
            fname = all_names.pop(n)
            self.controller.filelist.refresh(all_names)
            self.RemoveFile(fname)


    def onRemoveGroups(self, event=None):
        groups = []
        for checked in self.controller.filelist.GetCheckedStrings():
            groups.append(str(checked))
        if len(groups) < 1:
            return

        dlg = RemoveDialog(self, groups)
        res = dlg.GetResponse()
        dlg.Destroy()

        if res.ok:
            filelist = self.controller.filelist
            all_fnames = filelist.GetItems()
            for fname in groups:
                gname = self.controller.file_groups.pop(fname)
                delattr(self.controller.symtable, gname)
                all_fnames.remove(fname)

            filelist.Clear()
            for name in all_fnames:
                filelist.Append(name)
            self.controller.sync_xasgroups()

    def onFreezeGroups(self, event=None):
        self._freeze_handler(True)

    def onUnFreezeGroups(self, event=None):
        self._freeze_handler(False)

    def _freeze_handler(self, freeze):
        current_filename = self.current_filename
        reproc_group = None
        for fname in self.controller.filelist.GetCheckedStrings():
            groupname = self.controller.file_groups[str(fname)]
            dgroup = self.controller.get_group(groupname)
            if fname == current_filename:
                reproc_group = groupname
            dgroup.is_frozen = freeze
        if reproc_group is not None:
            self.ShowFile(groupname=reproc_group, process=True)

    def onMergeData(self, event=None):
        groups = {}
        for checked in self.controller.filelist.GetCheckedStrings():
            cname = str(checked)
            groups[cname] = self.controller.file_groups[cname]
        if len(groups) < 1:
            return

        outgroup = common_startstring(list(groups.keys()))
        if len(outgroup) < 2: outgroup = "data"
        outgroup = "%s (merge %d)" % (outgroup, len(groups))
        outgroup = unique_name(outgroup, self.controller.file_groups)
        dlg = MergeDialog(self, list(groups.keys()), outgroup=outgroup)
        res = dlg.GetResponse()
        dlg.Destroy()
        if res.ok:
            fname = res.group
            gname = fix_varname(res.group.lower())
            master = self.controller.file_groups[res.master]
            yname = 'norm' if res.ynorm else 'mu'
            this = self.controller.merge_groups(list(groups.values()),
                                                master=master,
                                                yarray=yname,
                                                outgroup=gname)

            mfiles, mgroups = [], []
            for g in groups.values():
                mgroups.append(g)
                mfiles.append(self.controller.get_group(g).filename)
            mfiles  = '[%s]' % (', '.join(mfiles))
            mgroups = '[%s]' % (', '.join(mgroups))
            desc = "%s: merge of %d groups" % (fname, len(groups))
            self.install_group(gname, fname, source=desc,
                               journal={'source_desc': desc,
                                        'merged_groups': mgroups,
                                        'merged_filenames': mfiles})

    def has_datagroup(self):
        return hasattr(self.controller.get_group(), 'energy')

    def onDeglitchData(self, event=None):
        if self.has_datagroup():
            DeglitchDialog(self, self.controller).Show()

    def onSmoothData(self, event=None):
        if self.has_datagroup():
            SmoothDataDialog(self, self.controller).Show()

    def onRebinData(self, event=None):
        if self.has_datagroup():
            RebinDataDialog(self, self.controller).Show()

    def onCorrectOverAbsorptionData(self, event=None):
        if self.has_datagroup():
            OverAbsorptionDialog(self, self.controller).Show()

    def onSpectraCalc(self, event=None):

        if self.has_datagroup():
            SpectraCalcDialog(self, self.controller).Show()

    def onEnergyCalibrateData(self, event=None):
        if self.has_datagroup():
            EnergyCalibrateDialog(self, self.controller).Show()

    def onDeconvolveData(self, event=None):
        if self.has_datagroup():
            DeconvolutionDialog(self, self.controller).Show()

    def onConfigDataFitting(self, event=None):
        pass

    def onAbout(self, event=None):
        info = AboutDialogInfo()
        info.SetName('Larix')
        info.SetDescription('X-ray Absorption Visualization and Analysis')
        info.SetVersion('Larch %s ' % larch.version.__version__)
        info.AddDeveloper('Matthew Newville: newville at cars.uchicago.edu')
        dlg = AboutBox(info)

    def onCheckforUpdates(self, event=None):
        dlg = LarchUpdaterDialog(self, caller='Larix')
        dlg.Raise()
        dlg.SetWindowStyle(wx.STAY_ON_TOP)
        res = dlg.GetResponse()
        dlg.Destroy()
        if res.ok and res.run_updates:
            from larch.apps import update_larch
            update_larch()
            self.onClose(event=event, prompt=False)

    def onClose(self, event=None, prompt=True):
        if prompt:
            dlg = QuitDialog(self, self.last_save_message)
            dlg.Raise()
            dlg.SetWindowStyle(wx.STAY_ON_TOP)
            res = dlg.GetResponse()
            dlg.Destroy()
            if not res.ok:
                return

        self.controller.save_workdir()
        try:
            self.controller.close_all_displays()
        except Exception:
            pass

        if self.larch_buffer is not None:
            try:
                self.larch_buffer.Destroy()
            except Exception:
                pass

        def destroy(wid):
            if hasattr(wid, 'Destroy'):
                try:
                    wid.Destroy()
                except Exception:
                    pass
                time.sleep(0.01)

        for name, wid in self.subframes.items():
            destroy(wid)

        for i in range(self.nb.GetPageCount()):
            nbpage = self.nb.GetPage(i)
            timers = getattr(nbpage, 'timers', None)
            if timers is not None:
                for t in timers.values():
                    t.Stop()

            if hasattr(nbpage, 'subframes'):
                for name, wid in nbpage.subframes.items():
                    destroy(wid)
        for t in self.timers.values():
            t.Stop()

        time.sleep(0.05)
        self.Destroy()

    def show_subframe(self, name, frameclass, **opts):
        shown = False
        if name in self.subframes:
            try:
                self.subframes[name].Raise()
                shown = True
            except:
                del self.subframes[name]
        if not shown:
            self.subframes[name] = frameclass(self, **opts)


    def onCIFBrowse(self, event=None):
        self.show_subframe('cif_feff', CIFFrame, _larch=self.larch,
                           path_importer=self.get_nbpage('feffit')[1].add_path,
                           with_feff=True)

    def onStructureBrowse(self, event=None):
        self.show_subframe('structure_feff', Structure2FeffFrame, _larch=self.larch,
                           path_importer=self.get_nbpage('feffit')[1].add_path)

    def onFeffBrowse(self, event=None):
        self.show_subframe('feff_paths', FeffResultsFrame, _larch=self.larch,
                           path_importer=self.get_nbpage('feffit')[1].add_path)

    def onLoadFitResult(self, event=None):
        pass
        # print("onLoadFitResult??")
        # self.nb.SetSelection(1)
        # self.nb_panels[1].onLoadFitResult(event=event)

    def onReadDialog(self, event=None):
        dlg = wx.FileDialog(self, message="Read Data File",
                            defaultDir=get_cwd(),
                            wildcard=FILE_WILDCARDS,
                            style=wx.FD_OPEN|wx.FD_MULTIPLE)
        self.paths2read = []
        if dlg.ShowModal() == wx.ID_OK:
            self.paths2read = dlg.GetPaths()
        dlg.Destroy()

        if len(self.paths2read) < 1:
            return

        def file_mtime(x):
            return os.stat(x).st_mtime

        self.paths2read = [Path(p).as_posix() for p in self.paths2read]
        self.paths2read = sorted(self.paths2read, key=file_mtime)

        path = self.paths2read.pop(0)

        do_read = True
        if path in self.controller.file_groups:
            do_read = (wx.ID_YES == Popup(self,
                                          "Re-read file '%s'?" % path,
                                          'Re-read file?'))
        if do_read:
            self.onRead(path)

    def onRead(self, path):
        fpath = Path(path).absolute()
        filedir = fpath.parent.as_posix()
        filename = fpath.name
        fullpath = fpath.as_posix()
        if self.controller.chdir_on_fileopen() and len(filedir) > 0:
            os.chdir(filedir)
            self.controller.set_workdir()

        # check for athena projects
        if is_athena_project(fullpath):
            self.show_subframe('athena_import', AthenaImporter,
                               controller=self.controller, filename=fullpath,
                               read_ok_cb=self.onReadAthenaProject_OK)
            return

        # check for Spec File
        if is_specfile(fullpath):
            self.show_subframe('spec_import', SpecfileImporter,
                               filename=fullpath,
                               _larch=self.larch_buffer.larchshell,
                               config=self.last_spec_config,
                               read_ok_cb=self.onReadSpecfile_OK)
            return

        # check for Larch Session File
        if is_larch_session_file(fullpath):
            self.onLoadSession(path=fullpath)
            return


        # default to Column File
        self.show_subframe('readfile', ColumnDataFileFrame, filename=fullpath,
                        config=self.last_col_config,
                        _larch=self.larch_buffer.larchshell,
                        read_ok_cb=self.onRead_OK)

    def onReadSpecfile_OK(self, script, path, scanlist, config=None):
        """read groups from a list of scans from a specfile"""
        self.larch.eval("_specfile = specfile('{path:s}')".format(path=path))
        dgroup = None
        fname = Path(path).name
        first_group = None
        cur_panel = self.nb.GetCurrentPage()
        cur_panel.skip_plotting = True
        symtable = self.larch.symtable
        if config is not None:
            self.last_spec_config = config

        array_desc = config.get('array_desc', {})

        multiconfig = config.get('multicol_config', {'channels':[], 'i0': config['iy2']})
        multi_i0  = multiconfig.get('i0', config['iy2'])
        multi_chans = copy.copy(multiconfig.get('channels', []))

        if len(multi_chans) > 0:
            if (multi_chans[0] == config['iy1'] and multi_i0 == config['iy2']
                and 'log' not in config['expressions']['yplot']):
                yname = config['array_labels'][config['iy1']]
                # filename = f"{spath}:{yname}"
                multi_chans.pop(0)

        for scan in scanlist:
            gname = fix_varname("{:s}{:s}".format(fname[:6], scan))
            if hasattr(symtable, gname):
                count, tname = 0, gname
                while count < 1e7 and self.larch.symtable.has_group(tname):
                    tname = gname + make_hashkey(length=7)
                    count += 1
                gname = tname

            cur_panel.skip_plotting = (scan == scanlist[-1])
            yname = config['yarr1']
            if first_group is None:
                first_group = gname
            cmd = script.format(group=gname, specfile='_specfile',
                                path=path, scan=scan, **config)

            self.larch.eval(cmd)
            displayname = f"{fname} scan{scan} {yname}"
            jrnl = {'source_desc': f"{fname}: scan{scan} {yname}"}
            dgroup = self.install_group(gname, displayname, journal=jrnl)
            if len(multi_chans) > 0:
                yplotline = None
                for line in script.split('\n'):
                    if line.startswith("{group}.yplot ="):
                        yplotline = line.replace("{group}", "{ngroup}")
                mscript = '\n'.join(["{ngroup} = deepcopy({group})",
                                     yplotline,
                                    "{ngroup}.mu = {ngroup}.yplot",
                                     "{ngroup}.plot_ylabel = '{ylabel}'"])
                i0 = '1.0'
                if multi_i0  < len(config['array_labels']):
                    i0 = config['array_labels'][multi_i0]

                for mchan in multi_chans:
                    yname = config['array_labels'][mchan]
                    ylabel = f"{yname}/{i0}"
                    dname = f"{fname} scan{scan} {yname}"
                    ngroup = file2groupname(dname, symtable=self.larch.symtable)
                    njournal = {'source': path,
                                'xplot': array_desc['xplot'].format(group=ngroup),
                                'yplot': ylabel,
                                'source_desc': f"{fname}: scan{scan} {yname}",
                                'yerr': array_desc['yerr'].format(group=ngroup)}
                    cmd = mscript.format(group=gname, ngroup=ngroup,
                                         iy1=mchan, iy2=multi_i0, ylabel=ylabel)
                    self.larch.eval(cmd)
                    self.install_group(ngroup, dname, source=path, journal=njournal)


        cur_panel.skip_plotting = False

        if first_group is not None:
            self.ShowFile(groupname=first_group, process=True, plot=True)
        self.write_message("read %d datasets from %s" % (len(scanlist), path))
        self.larch.eval('del _specfile')

    def onReadXasDataSource_OK(self, script, path, scanlist, array_sel=None, extra_sums=None):
        """read groups from a list of scans from a xas data source"""
        self.larch.eval("_data_source = open_xas_source('{path:s}')".format(path=path))
        dgroup = None
        fname = Path(path).name
        first_group = None
        cur_panel = self.nb.GetCurrentPage()
        cur_panel.skip_plotting = True
        symtable = self.larch.symtable
        if array_sel is not None:
            self.last_array_sel_spec = array_sel

        for scan in scanlist:
            gname = fix_varname("{:s}{:s}".format(fname[:6], scan))
            if hasattr(symtable, gname):
                count, tname = 0, gname
                while count < 1e7 and self.larch.symtable.has_group(tname):
                    tname = gname + make_hashkey(length=7)
                    count += 1
                gname = tname

            cur_panel.skip_plotting = (scan == scanlist[-1])
            yname = self.last_array_sel_spec['yarr1']
            if first_group is None:
                first_group = gname
            self.larch.eval(script.format(group=gname, data_source='_data_source',
                                          path=path, scan=scan))

            displayname = f"{fname:s} scan{scan:s} {yname:s}"
            jrnl = {'source_desc': f"{fname:s}: scan{scan:s} {yname:s}"}
            dgroup = self.install_group(gname, displayname,
                                        process=True, plot=False, extra_sums=extra_sums,
                                        source=displayname,
                                        journal=jrnl)
        cur_panel.skip_plotting = False

        if first_group is not None:
            self.ShowFile(groupname=first_group, process=True, plot=True)
        self.write_message("read %d datasets from %s" % (len(scanlist), path))
        self.larch.eval('del _data_source')

    def onReadAthenaProject_OK(self, path, namelist):
        """read groups from a list of groups from an athena project file"""
        self.larch.eval("_prj = read_athena('{path:s}', do_fft=False, do_bkg=False)".format(path=path))
        dgroup = None
        script = "{group:s} = _prj.{prjgroup:s}"
        cur_panel = self.nb.GetCurrentPage()
        cur_panel.skip_plotting = True
        parent, spath = path_split(path)
        labels = []
        groups_added = []

        for ig, gname in enumerate(namelist):
            cur_panel.skip_plotting = (gname == namelist[-1])
            this = getattr(self.larch.symtable._prj, gname)
            gid = file2groupname(str(getattr(this, 'athena_id', gname)),
                                 symtable=self.larch.symtable)
            if self.larch.symtable.has_group(gid):
                count, prefix = 0, gname[:3]
                while count < 1e7 and self.larch.symtable.has_group(gid):
                    gid = prefix + make_hashkey(length=7)
                    count += 1
            label = getattr(this, 'label', gname).strip()
            labels.append(label)

            jrnl = {'source_desc': f'{spath:s}: {gname:s}'}
            self.larch.eval(script.format(group=gid, prjgroup=gname))

            dgroup = self.install_group(gid, label, process=False,
                                        source=path, journal=jrnl)
            groups_added.append(gid)

        for gid in groups_added:
            rgroup = gid
            dgroup = self.larch.symtable.get_group(gid)

            conf_xasnorm = dgroup.config.xasnorm
            conf_exafs= dgroup.config.exafs

            apars = getattr(dgroup, 'athena_params', {})
            abkg = getattr(apars, 'bkg', {})
            afft = getattr(apars, 'fft', {})

            # norm
            for attr in ('e0', 'pre1', 'pre2', 'nnorm'):
                if hasattr(abkg, attr):
                    conf_xasnorm[attr] = float(getattr(abkg, attr))

            for attr, alt in (('norm1', 'nor1'), ('norm2', 'nor2'),
                              ('edge_step', 'step')):
                if hasattr(abkg, alt):
                    conf_xasnorm[attr]  = float(getattr(abkg, alt))
            if hasattr(abkg, 'fixstep'):
                a = float(getattr(abkg, 'fixstep', 0.0))
                conf_xasnorm['auto_step'] = (a < 0.5)


            # bkg
            for attr in ('e0', 'rbkg'):
                if hasattr(abkg, attr):
                    conf_exafs[attr] = float(getattr(abkg, attr))

            for attr, alt in (('bkg_kmin', 'spl1'), ('bkg_kmax', 'spl2'),
                              ('bkg_kweight', 'kw'), ('bkg_clamplo', 'clamp1'),
                              ('bkg_clamphi', 'clamp2')):
                if hasattr(abkg, alt):
                    val = getattr(abkg, alt)
                    try:
                        val = float(getattr(abkg, alt))
                    except:
                        if alt.startswith('clamp') and isinstance(val, str):
                            val = ATHENA_CLAMPNAMES.get(val.lower(), 0)
                    conf_exafs[attr] = val


            # fft
            for attr in ('kmin', 'kmax', 'dk', 'kwindow', 'kw'):
                if hasattr(afft, attr):
                    n = f'fft_{attr}'
                    if attr == 'kw': n = 'fft_kweight'
                    if attr == 'kwindow':
                        conf_exafs[n] = getattr(afft, attr)
                    else:
                        conf_exafs[n] = float(getattr(afft, attr))

            # reference
            refgroup = getattr(apars, 'reference', '')
            if refgroup in groups_added:
                newname = None
                for key, val in self.controller.file_groups.items():
                    if refgroup in (key, val):
                        newname = key

                if newname is not None:
                    refgroup = newname
            else:
                refgroup = dgroup.filename
            dgroup.energy_ref = refgroup

        self.larch.eval("del _prj")
        cur_panel.skip_plotting = False

        plot_first = True
        if len(labels) > 0:
            gname = self.controller.file_groups[labels[0]]
            self.ShowFile(groupname=gname, process=True, plot=plot_first)
            plot_first = False
        self.write_message("read %d datasets from %s" % (len(namelist), path))
        self.last_athena_file = path
        self.controller.sync_xasgroups()
        self.controller.recentfiles.append((time.time(), path))

    def onRead_OK(self, script, path, config):
        #groupname=None, filename=None,
        #          ref_groupname=None, ref_filename=None, config=None,
        #          array_desc=None):

        """ called when column data has been selected and is ready to be used
        overwrite: whether to overwrite the current datagroup, as when
        editing a datagroup
        """
        filedir, spath = path_split(path)
        filename = config.get('filename', spath)
        groupname = config.get('groupname', None)
        if groupname is None:
            groupname = file2groupname(filename,
                                       symtable=self.larch.symtable)
        array_desc = config.get('array_desc', {})
        if 'xplot' not in array_desc and 'xdat' in array_desc:  # back compat
            array_desc['xplot'] = copy.copy(array_desc['xdat'])
        if 'yplot' not in array_desc and 'ydat' in array_desc:  # back compat
            array_desc['yplot'] = copy.copy(array_desc['ydat'])

        if hasattr(self.larch.symtable, groupname):
            groupname = file2groupname(filename,
                                       symtable=self.larch.symtable)

        refgroup = config.get('refgroup', groupname + '_ref')

        multiconfig = config.get('multicol_config', {'channels':[], 'i0': config['iy2']})
        multi_i0  = multiconfig.get('i0', config['iy2'])
        multi_chans = copy.copy(multiconfig.get('channels', []))

        if len(multi_chans) > 0:
            if (multi_chans[0] == config['iy1'] and multi_i0 == config['iy2']
                and 'log' not in config['expressions']['yplot']):
                yname = config['array_labels'][config['iy1']]
                filename = f"{spath}:{yname}"
                multi_chans.pop(0)

        config = copy.copy(config)
        config['group'] = groupname
        config['path'] = path
        has_yref = config.get('has_yref', False)


        self.larch.eval(script.format(**config))

        if config is not None:
            self.last_col_config = config

        journal = {'source': path}
        refjournal = {}

        if 'xplot' in array_desc:
            journal['xplot'] = array_desc['xplot'].format(group=groupname)
        if 'yplot' in array_desc:
            journal['yplot'] = ylab = array_desc['yplot'].format(group=groupname)
            journal['source_desc'] = f'{spath}: {ylab}'
        if 'yerr' in array_desc:
            journal['yerr'] = array_desc['yerr'].format(group=groupname)

        self.install_group(groupname, filename, source=path, journal=journal)

        dtype = getattr(config, 'datatype', 'xydata')
        def install_multichans(config):
            yplotline = None
            yarray = 'mu' if dtype == 'xas' else 'y'
            for line in script.split('\n'):
                if line.startswith("{group}.yplot ="):
                    yplotline = line.replace("{group}", "{ngroup}")
            mscript = ["{ngroup} = deepcopy({group})",
                       yplotline,
                      "{ngroup}.{yarray} = {ngroup}.yplot[:]",
                      "{ngroup}.plot_ylabel = '{ylabel}'" ]
            if dtype == 'xydata':
                mscript.append("{ngroup}.scale = ptp({ngroup}.y+1.e-15)")

            i0 = '1.0'
            if multi_i0  < len(config['array_labels']):
                i0 = config['array_labels'][multi_i0]

            for mchan in multi_chans:
                yname = config['array_labels'][mchan]
                ylabel = f"{yname}/{i0}"
                fname = f"{spath}:{yname}"
                ngroup = file2groupname(fname, symtable=self.larch.symtable)
                njournal = {'source': path,
                            'xplot': array_desc['xplot'].format(group=ngroup),
                            'yplot': ylabel,
                            'source_desc': f"{spath}: {ylabel}",
                            'yerr': array_desc['yerr'].format(group=ngroup)}
                cmd = '\n'.join(mscript).format(group=config['group'],
                                                ngroup=ngroup, ylabel=ylabel,
                                                iy1=mchan, iy2=multi_i0,
                                                yarray=yarray)
                self.larch.eval(cmd)
                self.install_group(ngroup, fname, source=path, journal=njournal)

        if len(multi_chans) > 0:
            install_multichans(config)

        if has_yref:

            if 'xplot' in array_desc:
                refjournal['xplot'] = array_desc['xplot'].format(group=refgroup)
            if 'yref' in array_desc:
                refjournal['yplot'] = ydx = array_desc['yref'].format(group=refgroup)
                refjournal['source_desc'] = f'{spath:s}: {ydx:s}'
            self.install_group(refgroup, config['reffile'],
                               source=path, journal=refjournal)

        # check if rebin is needed
        thisgroup = getattr(self.larch.symtable, groupname)

        do_rebin = False
        if thisgroup.datatype == 'xas':
            try:
                en = thisgroup.energy
            except:
                do_rebin = True
                en = thisgroup.energy = thisgroup.xplot
            # test for rebinning:
            #  too many data points
            #  unsorted energy data or data in angle
            #  too fine a step size at the end of the data range
            if (len(en) > 1200 or
                any(np.diff(en) < 0) or
                ((max(en)-min(en)) > 300 and
                 (np.diff(en[-50:]).mean() < 0.75))):
                msg = """This dataset may need to be rebinned.
                Rebin now?"""
                dlg = wx.MessageDialog(self, msg, 'Warning',
                                       wx.YES | wx.NO )
                do_rebin = (wx.ID_YES == dlg.ShowModal())
                dlg.Destroy()
        gname = None

        for path in self.paths2read:
            filedir, spath = path_split(path)
            fname = spath
            if len(multi_chans) > 0:
                yname = config['array_labels'][config['iy1']]
                fname = f"{spath}:{yname}"

            gname = file2groupname(fname, symtable=self.larch.symtable)
            refgroup = config['refgroup']
            if has_yref:
                refgroup = gname + '_ref'
                reffile = spath + '_ref'
            config = copy.copy(config)
            config['group'] = gname
            config['refgroup'] = refgroup
            config['path'] = path

            self.larch.eval(script.format(**config))
            if has_yref:
                self.larch.eval(f"{gname}.energy_ref = {refgroup}.energy_ref = '{refgroup}'\n")

            if 'xplot' in array_desc:
                journal['xplot'] = array_desc['xplot'].format(group=gname)
            if 'yplot' in array_desc:
                journal['yplot'] = ydx = array_desc['yplot'].format(group=gname)
                journal['source_desc'] = f'{spath:s}: {ydx:s}'
            if 'yerr' in array_desc:
                journal['yerr'] = array_desc['yerr'].format(group=gname)

            self.install_group(gname, fname, source=path, journal=journal, plot=False)
            if len(multi_chans) > 0:
                install_multichans(config)

            if has_yref:
                if 'xplot' in array_desc:
                    refjournal['xplot'] = array_desc['xplot'].format(group=refgroup)
                if 'yref' in array_desc:
                    refjournal['yplot'] = ydx = array_desc['yref'].format(group=refgroup)
                    refjournal['source_desc'] = f'{spath:s}: {ydx:s}'

                self.install_group(refgroup, reffile, source=path, journal=refjournal, plot=False)


        if gname is not None:
            self.ShowFile(groupname=gname)

        self.write_message("read %s" % (spath))
        if do_rebin:
            RebinDataDialog(self, self.controller).Show()

    def install_group(self, groupname, filename=None, source=None, journal=None,
                      process=True, plot=True):
        """add groupname / filename to list of available data groups"""
        if isinstance(groupname, Group):
            dgroup = groupname
            groupname = groupname.groupname
        else:
            dgroup = self.controller.get_group(groupname)

        if filename is None:
            filename = dgroup.filename

        self.controller.install_group(groupname, filename,
                                      source=source, journal=journal)
        dtype = getattr(dgroup, 'datatype', 'xydata')
        startpage = 'xasnorm' if dtype == 'xas' else 'xydata'
        ipage, pagepanel = self.get_nbpage(startpage)
        self.nb.SetSelection(ipage)
        self.ShowFile(groupname=groupname, filename=filename,
                      process=process, plot=plot)

    ##
    def get_recent_session_menu(self):
        """ get recent sessions files for Menu list"""
        for menu_item in self.recent_menu.MenuItems:
            self.recent_menu.Remove(menu_item)

        for tstamp, fname in self.controller.get_recentfiles():
            message =  f"{fname} [{time_ago(tstamp)} ago]"
            MenuItem(self, self.recent_menu, message,
                     f"file saved {isotime(tstamp)}",
                     partial(self.onLoadSession, path=fname))

        self.recent_menu.AppendSeparator()
        for tstamp, fname in self.controller.recent_autosave_sessions():
            message =  f"{fname} [{time_ago(tstamp)} ago]"
            if abs(tstamp) < 5.0:
                message =  f"{fname} [most recent]"
            MenuItem(self, self.recent_menu, message,
                     f"file saved {isotime(tstamp)}",
                     partial(self.onLoadSession, path=fname))

    def onAutoSaveTimer(self, event=None):
        """autosave session periodically, using autosave_config settings
        and avoiding saving sessions while program is inactive.
        """
        conf = self.controller.get_config('autosave', {})
        savetime = conf.get('savetime', 600)
        symtab = self.larch.symtable

        if (time.time() > self.last_autosave + savetime and
            symtab._sys.last_eval_time > (self.last_autosave+60) and
            len(symtab._xasgroups) > 0):
            self.autosave_session()
            self.get_recent_session_menu()

    def autosave_session(self, event=None):
        """autosave session now"""
        savefile = self.controller.autosave_session()
        # save_session(savefile, _larch=self.larch._larch)
        self.last_autosave = time.time()
        stime = time.strftime("%H:%M")
        self.last_save_message = ("Session last saved", f"'{savefile}'", f"{stime}")
        self.write_message(f"Session saved to '{savefile}' at {stime}")

    ## float-spin / pin timer events
    def onPinTimer(self, event=None):
        if 'start' not in self.cursor_dat:
            self.cursor_dat['xsel'] = None
            self.onPinTimerComplete(reason="bad")
        pin_config = self.controller.get_config('pin',
                                                {'style': 'pin_first',
                                                 'max_time':15.0,
                                                 'min_time': 2.0})
        min_time = float(pin_config['min_time'])
        timeout = float(pin_config['max_time'])

        curhist_name = self.cursor_dat['name']
        cursor_hist = getattr(self.larch.symtable._plotter, curhist_name, [])
        if len(cursor_hist) > self.cursor_dat['nhist']: # got new data!
            self.cursor_dat['xsel'] = cursor_hist[0][0]
            self.cursor_dat['ysel'] = cursor_hist[0][1]
            if time.time() > min_time + self.cursor_dat['start']:
                self.timers['pin'].Stop()
                self.onPinTimerComplete(reason="new")
        elif time.time() > timeout + self.cursor_dat['start']:
            self.onPinTimerComplete(reason="timeout")

        if 'win' in self.cursor_dat and 'xsel' in self.cursor_dat:
            time_remaining = timeout + self.cursor_dat['start'] - time.time()
            msg = 'Select Point from Plot #%d' % (self.cursor_dat['win'])
            if self.cursor_dat['xsel'] is not None:
                msg = '%s, [current value=%.1f]' % (msg, self.cursor_dat['xsel'])
            msg = '%s, expiring in %.0f sec' % (msg, time_remaining)
            self.write_message(msg)

    def onPinTimerComplete(self, reason=None, **kws):
        self.timers['pin'].Stop()
        if reason != "bad":
            msg = 'Selected Point at %.1f' % self.cursor_dat['xsel']
            if reason == 'timeout':
                msg = msg + '(timed-out)'
            self.write_message(msg)
            if (self.cursor_dat['xsel'] is not None and
                callable(self.cursor_dat['callback'])):
                self.cursor_dat['callback'](**self.cursor_dat)
                time.sleep(0.05)
        else:
            self.write_message('Select Point Error')
        self.cursor_dat = {}


    def onSelPoint(self, evt=None, opt='__', relative_e0=True, callback=None,
                   win=None):
        """
        get last selected point from a specified plot window
        and fill in the value for the widget defined by `opt`.

        start Pin Timer to get last selected point from a specified plot window
        and fill in the value for the widget defined by `opt`.
        """
        if win is None:
            win = 1

        display = get_display(win=win, _larch=self.larch)
        display.Raise()
        msg = 'Select Point from Plot #%d' % win
        self.write_message(msg)

        now = time.time()
        curhist_name = 'plot%d_cursor_hist' % win
        cursor_hist = getattr(self.larch.symtable._plotter, curhist_name, [])

        self.cursor_dat = dict(relative_e0=relative_e0, opt=opt,
                               callback=callback,
                               start=now, xsel=None, ysel=None,
                               win=win, name=curhist_name,
                               nhist=len(cursor_hist))

        pin_config = self.controller.get_config('pin',
                                                {'style': 'pin first',
                                                 'max_time':15.0,
                                                 'min_time': 2.0})
        if pin_config['style'].startswith('plot'):
            if len(cursor_hist) > 0:
                x, y, t = cursor_hist[0]
                if now < (t + 60.0):
                    self.cursor_dat['xsel'] = x
                    self.cursor_dat['ysel'] = y
                    msg = 'Selected Point at x=%.3f' % self.cursor_dat['xsel']
                    self.cursor_dat['callback'](**self.cursor_dat)
            else:
                self.write_message('No Points selected from plot window!')
        else: # "pin first" mode
            if len(cursor_hist) > 2:  # purge old cursor history
                setattr(self.larch.symtable._plotter, curhist_name, cursor_hist[:2])

            if len(cursor_hist) > 0:
                x, y, t = cursor_hist[0]
                if now < (t + 30.0):
                    self.cursor_dat['xsel'] = x
                    self.cursor_dat['ysel'] = y
            self.timers['pin'].Start(250)


class LarixApp(LarchWxApp):
    def __init__(self, filename=None, check_version=True, mode='xas',
                 with_wx_inspect=False, **kws):
        self.filename = filename
        self.mode = mode
        self.with_wx_inspect = with_wx_inspect
        self.check_version = check_version
        LarchWxApp.__init__(self,**kws)

    def createApp(self):
        self.frame = LarixFrame(filename=self.filename,
                                mode=self.mode,
                                with_wx_inspect=self.with_wx_inspect,
                                check_version=self.check_version)
        self.SetTopWindow(self.frame)
        # if self.with_wx_inspect:
        #    wx.GetApp().ShowInspectionTool()
        return True

def larix(**kws):
    LarixApp(**kws)

if __name__ == "__main__":
    import argparse
    parser = argparse.ArgumentParser(description=LARIX_TITLE)
    parser.add_argument(
        '-f', '--filename',
        dest='filename',
        help='data file to load')
    parser.add_argument(
        '-m', '--mode',
        dest='mode',
        help='mode to start larix')
    parser.add_argument(
        '-w', '--wx_inspect',
        dest='wx_inspect',
        help='wx debugging mode')
    args = parser.parse_args()
    LarixApp(**vars(args)).MainLoop()



from docutils import nodes
from sphinx.roles import XRefRole
import figtable
import subfig
from backports import OrderedDict, OrderedSet

# Element classes

class page_ref(nodes.reference):
    pass

class num_ref(nodes.reference):
    pass


# Visit/depart functions

def skip_page_ref(self, node):
    raise nodes.SkipNode

def latex_visit_page_ref(self, node):
    self.body.append("\\pageref{%s:%s}" % (node['refdoc'], node['reftarget']))
    raise nodes.SkipNode

def latex_visit_num_ref(self, node):
    fields = node['reftarget'].split('#')

    if len(fields) > 1:
        label, target = fields
    else:
        label = None
        target = fields[0]

    if target not in self.builder.env.docnames_by_figname:
        raise nodes.SkipNode
    targetdoc = self.builder.env.docnames_by_figname[target]

    ref_link = '%s:%s' % (targetdoc, target)

    if label is None:
        latex = '\\ref{%s}' % ref_link
    else:
        latex = "\\hyperref[%s]{%s \\ref*{%s}}" % (ref_link, label, ref_link)

    self.body.append(latex)
    raise nodes.SkipNode


def doctree_read(app, doctree):
    # first generate figure numbers for each figure
    env = app.builder.env

    docname_figs = getattr(env, 'docname_figs', {})
    docnames_by_figname = getattr(env, 'docnames_by_figname', {})

    for figure_info in doctree.traverse(lambda n: isinstance(n, nodes.figure) or \
                                                  isinstance(n, subfig.subfigend) or \
                                                  isinstance(n, figtable.figtable)):

        for id in figure_info['ids']:

            fig_docname = docnames_by_figname[id] = env.docname

            if fig_docname not in docname_figs:
                docname_figs[fig_docname] = OrderedDict()

            if isinstance(figure_info.parent, subfig.subfig):
                mainid = figure_info.parent['mainfigid']
            else:
                mainid = id

            if mainid not in docname_figs[fig_docname]:
                docname_figs[fig_docname][mainid] = OrderedSet()

            if isinstance(figure_info.parent, subfig.subfig):
                docname_figs[fig_docname][mainid].add(id)

    env.docnames_by_figname = docnames_by_figname
    env.docname_figs = docname_figs

def figname_is_id(s):
    "determine if string is 'idX' where X is an integer"
    if s.startswith('id') and len(s) > 2:
        isint = False
        try:
            isint = int(s[2:]) is not None
        except ValueError:
            isint = False
        return isint

def doctree_resolved(app, doctree, docname):
    # replace numfig nodes with links
    if app.builder.name in ('html', 'singlehtml', 'epub'):
        env = app.builder.env

        docname_figs = getattr(env, 'docname_figs', {})
        docnames_by_figname = env.docnames_by_figname

        figids = getattr(env, 'figids', {})

        secnums = []
        fignames_by_secnum = {}
        for figdocname, figurelist in env.docname_figs.iteritems():
            if figdocname not in env.toc_secnumbers:
                continue
            secnum = env.toc_secnumbers[figdocname]['']
            secnums.append(secnum)
            fignames_by_secnum[secnum] = figurelist

        last_secnum = 0
        secnums = sorted(secnums)
        figid = 1
        istrs = [str(i) for i in range(10)]
        for secnum in secnums:
            if secnum[0] != last_secnum:
                figid = 1
            for figname, subfigs in fignames_by_secnum[secnum].iteritems():
                figids[figname] = str(secnum[0]) + '.' + str(figid)
                isub = 0
                for subfigname in subfigs:
                    subfigid = figids[figname] + chr(ord('a') + isub)
                    figids[subfigname] = subfigid
                    if not figname_is_id(subfigname):
                        isub += 1
                if not figname_is_id(figname):
                    figid += 1
            last_secnum = secnum[0]

            env.figids = figids

        for figure_info in doctree.traverse(lambda n: isinstance(n, nodes.figure) or \
                                                      isinstance(n, subfig.subfigend) or \
                                                      isinstance(n, figtable.figtable)):


            try:
                id = figure_info['ids'][0]
                fignum = figids[id]
            except (IndexError, KeyError):
                continue

            for cap in figure_info.traverse(nodes.caption):
                cap.insert(1, nodes.Text(" %s" % cap[0]))
                print(" CAPTION ", cap, fignum[-1])
                if fignum[-1] in istrs:
                    boldcaption = "%s %s:" % (app.config.figure_caption_prefix, fignum)
                else:
                    boldcaption = "(%s)" % fignum[-1]
                cap[0] = nodes.strong('', boldcaption)
                print(" --> ", boldcaption, cap)


        for ref_info in doctree.traverse(num_ref):
            if '#' in ref_info['reftarget']:
                label, target = ref_info['reftarget'].split('#')
                labelfmt = label + " %s"
            else:
                labelfmt = '%s'
                target = ref_info['reftarget']

            if target not in docnames_by_figname:
                app.warn('Target figure not found: %s' % target)
                link = "#%s" % target
                linktext = target
            else:
                target_doc = docnames_by_figname[target]

                if app.builder.name == 'singlehtml':
                    link = "#%s" % target
                else:
                    link = "%s#%s"%(app.builder.get_relative_uri(docname,
                                                                 target_doc),
                                    target)
                if target in figids:
                    linktext = labelfmt % figids[target]
                else:
                    linktext = target


            html = '<a href="%s">%s</a>' % (link, linktext)
            ref_info.replace_self(nodes.raw(html, html, format='html'))

def setup(app):
    app.add_config_value('number_figures', True, True)
    app.add_config_value('figure_caption_prefix', "Figure", True)

    app.add_node(page_ref,
                 text=(skip_page_ref, None),
                 html=(skip_page_ref, None),
                 singlehtml=(skip_page_ref, None),
                 latex=(latex_visit_page_ref, None))

    app.add_role('page', XRefRole(nodeclass=page_ref))

    app.add_node(num_ref,
                 latex=(latex_visit_num_ref, None),
                 text=(skip_page_ref, None))

    app.add_role('num', XRefRole(nodeclass=num_ref))

    app.connect('doctree-read', doctree_read)
    app.connect('doctree-resolved', doctree_resolved)



#!/usr/bin/env pythonw
'''
GUI for displaying 1D XRD images

'''
import os
import sys
import time

from functools import partial

import numpy as np
from numpy.polynomial.chebyshev import chebfit, chebval

from pyFAI.azimuthalIntegrator import AzimuthalIntegrator
import pyFAI.units
pyFAI.use_opencl = False

import wx
import wx.lib.scrolledpanel as scrolled
from wxmplot import PlotPanel

from lmfit.lineshapes import gaussian

import larch
from larch.larchlib import read_workdir, save_workdir
from larch.utils import get_cwd, gformat, fix_filename
from larch.utils.physical_constants import PLANCK_HC
from larch.xray import XrayBackground
from larch.wxxas.xas_dialogs import RemoveDialog
from larch.io import tifffile

from larch.xrd import (d_from_q,twth_from_q,q_from_twth,
                       d_from_twth,twth_from_d,q_from_d,
                       lambda_from_E, E_from_lambda, calc_broadening,
                       instrumental_fit_uvw,peaklocater,peakfitter,
                       xrd1d, save1D, read_poni)

from larch.wxlib import (ReportFrame, BitmapButton, FloatCtrl, FloatSpin,
                         SetTip, GridPanel, get_icon, SimpleText, pack,
                         Button, HLine, Choice, Check, MenuItem, COLORS,
                         set_color, CEN, RIGHT, LEFT, FRAMESTYLE, Font,
                         FONTSIZE, FONTSIZE_FW, FileSave, FileOpen,
                         flatnotebook, Popup, FileCheckList, OkCancel,
                         EditableListBox, ExceptionPopup, CIFFrame,
                         LarchFrame, LarchWxApp)

MAXVAL = 2**32 - 2**15
MAXVAL_INT16 = 2**16 - 8

XYWcards = "XY Data File(*.xy)|*.xy|All files (*.*)|*.*"
TIFFWcards = "TIFF Files|*.tif;*.tiff|All files (*.*)|*.*"
PlotWindowChoices = ['1', '2', '3', '4', '5', '6', '7', '8', '9']

X_SCALES = [u'q (\u212B\u207B\u00B9)', u'2\u03B8 (\u00B0)', u'd (\u212B)']
Y_SCALES = ['linear', 'log']

PLOT_TYPES = {'Raw Data': 'raw',
              'Raw Data + Background' : 'raw+bkg',
              'Background-subtracted Data': 'sub'}

PLOT_CHOICES = list(PLOT_TYPES.keys())
PLOT_CHOICES_MULTI = [PLOT_CHOICES[0], PLOT_CHOICES[2]]

SCALE_METHODS = {'Max Raw Intensity': 'raw_max',
                 'Max Background-Subtracted Intensity': 'sub_max',
                 'Max Background Intensity': 'bkg_max',
                 'Mean Raw Intensity': 'raw_mean',
                 'Mean Background-Subtracted Intensity': 'sub_mean',
                 'Mean Background Intensity': 'bkg_mean'}

def keyof(dictlike, value, default):
    "dict reverse lookup"
    if value not in dictlike.values():
        value = default
    defout = None
    for key, val in dictlike.items():
        if defout is None:
            defout = key
        if val == value:
            return key
    return defout


def smooth_bruckner(y, smooth_points, iterations):
    y_original = y
    N_data = y.size
    N = smooth_points
    N_float = float(N)
    y = np.empty(N_data + N + N)

    y[0:N].fill(y_original[0])
    y[N:N + N_data] = y_original[0:N_data]
    y[N + N_data:N_data + N + N].fill(y_original[-1])

    y_avg = np.average(y)
    y_min = np.min(y)

    y_c = y_avg + 2. * (y_avg - y_min)
    y[y > y_c] = y_c

    window_size = N_float*2+1

    for j in range(0, iterations):
        window_avg = np.average(y[0: 2*N + 1])
        for i in range(N, N_data - 1 - N - 1):
            if y[i]>window_avg:
                y_new = window_avg
                #updating central value in average (first bracket)
                #and shifting average by one index (second bracket)
                window_avg += ((window_avg-y[i]) + (y[i+N+1]-y[i - N]))/window_size
                y[i] = y_new
            else:
                #shifting average by one index
                window_avg += (y[i+N+1]-y[i - N])/window_size
    return y[N:N + N_data]

def extract_background(x, y, smooth_width=0.1, iterations=40, cheb_order=40):
    """DIOPTAS
    Performs a background subtraction using bruckner smoothing and a chebyshev polynomial.
    Standard parameters are found to be optimal for synchrotron XRD.
    :param x: x-data of pattern
    :param y: y-data of pattern
    :param smooth_width: width of the window in x-units used for bruckner smoothing
    :param iterations: number of iterations for the bruckner smoothing

    :param cheb_order: order of the fitted chebyshev polynomial
    :return: vector of extracted y background
    """
    smooth_points = int((float(smooth_width) / (x[1] - x[0])))
    y_smooth = smooth_bruckner(y, abs(smooth_points), iterations)
    # get cheb input parameters
    x_cheb = 2. * (x - x[0]) / (x[-1] - x[0]) - 1.
    cheb_params = chebfit(x_cheb, y_smooth, cheb_order)
    return chebval(x_cheb, cheb_params)

def calc_bgr(dset, qwid=0.1, nsmooth=40, cheb_order=40):
    try:
        bgr = extract_background(dset.q, dset.I,
                                 smooth_width=qwid,
                                 iterations=nsmooth,
                                 cheb_order=cheb_order)
    except:
        bgr = 0.0*dset.I
    return bgr


class WavelengthDialog(wx.Dialog):
    """dialog for wavelength/energy"""
    def __init__(self, parent, wavelength, callback=None):

        self.parent = parent
        self.callback = callback

        wx.Dialog.__init__(self, parent, wx.ID_ANY, size=(550, 400),
                           title="Set Wavelength / Energy")
        self.SetFont(Font(FONTSIZE))
        panel = GridPanel(self, ncols=3, nrows=4, pad=4, itemstyle=LEFT)

        self.wids = wids = {}

        opts  = dict(size=(90, -1), act_on_losefocus=True)
        wids['wavelength'] = FloatCtrl(panel, value=wavelength, precision=7,
                                       minval=1.0e-4, maxval=100, **opts)

        en_ev = PLANCK_HC/wavelength
        wids['energy'] = FloatCtrl(panel, value=en_ev, precision=2,
                                   minval=50, maxval=5.e5, **opts)


        wids['wavelength'].SetAction(self.set_wavelength)
        wids['energy'].SetAction(self.set_energy)

        panel.Add(SimpleText(panel, 'Wavelength(\u212B): '),
                  dcol=1, newrow=False)
        panel.Add(wids['wavelength'], dcol=1)
        panel.Add(SimpleText(panel, 'Energy (eV): '),
                  dcol=1, newrow=True)
        panel.Add(wids['energy'], dcol=1)

        panel.Add((10, 10), newrow=True)

        panel.Add(Button(panel, 'Done', size=(150, -1),
                         action=self.onDone),  newrow=True)
        panel.pack()
        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add(panel, 1, LEFT, 5)
        pack(self, sizer)
        self.Fit()

        w0, h0 = self.GetSize()
        w1, h1 = self.GetBestSize()
        self.SetSize((max(w0, w1)+25, max(h0, h1)+25))

    def onDone(self, event=None):
        if callable(self.callback):
            self.callback(self.wids['wavelength'].GetValue())
        self.Destroy()

    def set_wavelength(self, value=1, event=None):
        w = self.wids['wavelength'].GetValue()

        self.wids['energy'].SetValue(PLANCK_HC/w, act=False)

    def set_energy(self, value=10000, event=None):
        w = self.wids['energy'].GetValue()
        self.wids['wavelength'].SetValue(PLANCK_HC/w, act=False)

class RenameDialog(wx.Dialog):
    """dialog for renaming a pattern"""
    def __init__(self, parent, name, callback=None):

        self.parent = parent
        self.callback = callback

        wx.Dialog.__init__(self, parent, wx.ID_ANY, size=(550, 400),
                           title="Rename dataset")
        self.SetFont(Font(FONTSIZE))
        panel = GridPanel(self, ncols=3, nrows=4, pad=4, itemstyle=LEFT)

        self.wids = wids = {}
        wids['newname'] = wx.TextCtrl(panel, value=name, size=(150, -1))

        panel.Add(SimpleText(panel, 'New Name: '),  dcol=1, newrow=False)
        panel.Add(wids['newname'], dcol=1)
        panel.Add((10, 10), newrow=True)

        panel.Add(Button(panel, 'Done', size=(150, -1),
                         action=self.onDone),  newrow=True)
        panel.pack()
        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add(panel, 1, LEFT, 5)
        pack(self, sizer)
        self.Fit()

        w0, h0 = self.GetSize()
        w1, h1 = self.GetBestSize()
        self.SetSize((max(w0, w1)+25, max(h0, h1)+25))

    def onDone(self, event=None):
        if callable(self.callback):
            self.callback(self.wids['newname'].GetValue())
        self.Destroy()

class ResetMaskDialog(wx.Dialog):
    """dialog for wavelength/energy"""
    def __init__(self, parent):
        wx.Dialog.__init__(self, parent, wx.ID_ANY, size=(350, 300),
                           title="Unset Mask?")
        self.SetFont(Font(FONTSIZE))
        panel = GridPanel(self, ncols=3, nrows=4, pad=4, itemstyle=LEFT)

        self.wids = wids = {}

        warn_msg = 'This will remove the current mask!'

        panel.Add(SimpleText(panel, warn_msg), dcol=2)

        panel.Add(OkCancel(panel), dcol=2, newrow=True)
        panel.pack()

        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add(panel, 1, LEFT, 5)
        pack(self, sizer)
        self.Fit()
        w0, h0 = self.GetSize()
        w1, h1 = self.GetBestSize()
        self.SetSize((max(w0, w1)+25, max(h0, h1)+25))

    def GetResponse(self):
        self.Raise()
        return (self.ShowModal() == wx.ID_OK)

class XRD1DFrame(wx.Frame):
    """browse 1D XRD patterns"""

    def __init__(self, parent=None, wavelength=1.0, ponifile=None,
                 _larch=None, **kws):

        wx.Frame.__init__(self, None, -1, title='1D XRD Browser',
                          style=FRAMESTYLE, size=(600, 600), **kws)
        self.parent = parent
        self.wavelength = wavelength
        self.poni = {'wavelength': 1.e-10*self.wavelength} # ! meters!
        self.pyfai_integrator = None

        self.larch = _larch
        if self.larch is None:
            self.larch_buffer = LarchFrame(_larch=None, parent=self,
                                           is_standalone=False,
                                           with_raise=False,
                                           exit_on_close=False)

            self.larch = self.larch_buffer.larchshell

        self.current_label = None
        self.cif_browser = None
        self.img_display = None
        self.plot_display = None
        self.mask = None
        self.datasets = {}
        self.form = {}
        self.createMenus()
        self.build()
        self.set_wavelength(self.wavelength)
        if ponifile is not None:
            self.set_ponifile(ponifile)

    def createMenus(self):
        fmenu = wx.Menu()
        cmenu = wx.Menu()
        smenu = wx.Menu()
        MenuItem(self, fmenu, "Read XY File",
                 "Read XRD 1D data from XY FIle",
                 self.onReadXY)

        MenuItem(self, fmenu, "Save XY File",
                 "Save XRD 1D data to XY FIle",
                 self.onSaveXY)
        self.tiff_reader = MenuItem(self, fmenu, "Read TIFF XRD Image",
                                    "Read XRD 2D image to be integrated",
                                    self.onReadTIFF)
        self.tiff_reader.Enable(self.poni.get('dist', -1) > 0)
        fmenu.AppendSeparator()
        MenuItem(self, fmenu, "Change Label for Current Pattern",
                 "Rename Current Pattern",
                 self.onRenameDataset)

        MenuItem(self, fmenu, "Remove Selected Patterns",
                 "Remove Selected Patterns",
                 self.remove_selected_datasets)

        fmenu.AppendSeparator()
        MenuItem(self, fmenu, "&Quit\tCtrl+Q", "Quit program", self.onClose)

        MenuItem(self, smenu, "Browse AmMin Crystal Structures",
                 "Browse Structures from American Mineralogical Database",
                 self.onCIFBrowse)

        MenuItem(self, cmenu, "Read PONI Calibration File",
                 "Read PONI Calibration (pyFAI) File",
                 self.onReadPONI)

        MenuItem(self, cmenu, "Set Energy/Wavelength",
                 "Set Energy and Wavelength",
                 self.onSetWavelength)

        MenuItem(self, cmenu, "Set Mask for imported Images",
                 "Read Mask for Imported TIFF XRD Images", self.onReadMask)

        m = MenuItem(self, cmenu, "Unset Mask",
                     "Reset to use no mask for Imported TIFF XRD Images",
                     self.onUnsetMask)
        self.unset_mask_menu = m
        m.Enable(False)

        m = MenuItem(self, cmenu, "Show Mask Image",
                     "Show image of mask", self.onShowMask)
        self.show_mask_menu = m
        m.Enable(False)

        menubar = wx.MenuBar()
        menubar.Append(fmenu, "&File")
        menubar.Append(cmenu, "&Calibration and Mask")
        menubar.Append(smenu, "&Search CIF Structures")
        self.SetMenuBar(menubar)


    def onClose(self, event=None):
        try:
            if self.panel is not None:
                self.panel.win_config.Close(True)
            if self.panel is not None:
                self.panel.win_config.Destroy()
        except:
            pass

        for attr in ('cif_browser', 'img_display', 'plot_display'):
            winx = getattr(self, attr, None)
            if winx is not None:
                try:
                    winx.Destroy()
                except:
                    pass

        if hasattr(self.larch.symtable, '_plotter'):
            wx.CallAfter(self.larch.symtable._plotter.close_all_displays)

        self.Destroy()

    def onSetWavelength(self, event=None):
        WavelengthDialog(self, self.wavelength, self.set_wavelength).Show()

    def onReadPONI(self, event=None):
        sfile = FileOpen(self, 'Read PONI (pyFAI) calibration file',
                         default_file='XRD.poni',
                         default_dir=get_cwd(),
                         wildcard="PONI Files(*.poni)|*.poni|All files (*.*)|*.*")

        if sfile is not None:
            try:
                self.set_poni(read_poni(sfile), with_pyfai=True)
            except:
                title = "Could not read PONI File"
                message = [f"Could not read PONI file {sfile}"]
                ExceptionPopup(self, title, message)

            top, xfile = os.path.split(sfile)
            os.chdir(top)

            if self.pyfai_integrator is None:
                try:
                    self.pyfai_integrator = AzimuthalIntegrator(**self.poni)
                except:
                    self.pyfai_integrator = None

            self.tiff_reader.Enable(self.pyfai_integrator is not None)

        self.set_wavelength(self.poni['wavelength']*1.e10)

    def onReadXY(self, event=None):
        sfile = FileOpen(self, 'Read XY Data',
                         default_file='XRD.xy',
                         default_dir=get_cwd(),
                         wildcard=XYWcards)
        if sfile is not None:
            top, xfile = os.path.split(sfile)
            os.chdir(top)
            dxrd = xrd1d(file=sfile, wavelength=self.wavelength)
            self.add_data(dxrd, label=xfile)

    def onUnsetMask(self, event=None):
        if self.mask is not None:
            dlg = ResetMaskDialog(self)
            if dlg.GetResponse():
                self.mask = None
                self.unset_mask_menu.Enable(False)
                self.show_mask_menu.Enable(False)

    def onReadMask(self, event=None):
        sfile = FileOpen(self, 'Read Mask Image File',
                         default_file='XRD.mask',
                         default_dir=get_cwd(),
                         wildcard="Mask Files(*.mask)|*.mask|All files (*.*)|*.*")

        if sfile is not None:
            valid_mask = False
            try:
                img =  tifffile.imread(sfile)
                valid_mask = len(img.shape)==2 and img.max() == 1 and img.min() == 0
            except:
                valid_mask = False
            if valid_mask:
                self.mask = (1 - img[::-1, :]).astype(img.dtype)
                self.unset_mask_menu.Enable(True)
                self.show_mask_menu.Enable(True)
            else:
                title = "Could not use mask file"
                message = [f"Could not use {sfile:s} as a mask file"]
                o = ExceptionPopup(self, title, message)

    def onShowMask(self, event=None):
        if self.mask is not None:
            imd = self.get_imdisplay()
            imd.display(self.mask, colomap='gray', auto_contrast=True)

    def onReadTIFF(self, event=None):
        sfile = FileOpen(self, 'Read TIFF XRD Image',
                         default_file='XRD.tiff',
                         default_dir=get_cwd(),
                         wildcard=TIFFWcards)
        if sfile is not None:
            top, fname = os.path.split(sfile)            
            
            if self.pyfai_integrator is None:
                try:
                    self.pyfai_integrator = AzimuthalIntegrator(**self.poni)
                except:
                    title = "Could not create pyFAI integrator: bad PONI data?"
                    message = [f"Could not create pyFAI integrator"]
                    ExceptionPopup(self, title, message)
            if self.pyfai_integrator is None:
                return

            img =  tifffile.imread(sfile)
            self.display_xrd_image(img, label=fname)

    def display_xrd_image(self, img, label='Image'):
        if self.mask is not None:
            if (self.mask.shape == img.shape):
                img = img*self.mask
            else:
                title = "Could not apply current mask"
                message = [f"Could not apply current mask [shape={self.mask.shape}]",
                           f"to this XRD image [shape={img.shape}]"]
                o = ExceptionPopup(self, title, message)

        if (img.max() > MAXVAL_INT16) and (img.max() < MAXVAL_INT16 + 64):
            #probably really 16bit data
            img[np.where(img>MAXVAL_INT16)] = 0
        else:
            img[np.where(img>MAXVAL)] = 0
        img[np.where(img<-1)] = -1
        img = img[::-1, :]

        imd = self.get_imdisplay()
        imd.display(img, colomap='gray', auto_contrast=True)
            
        integrate = self.pyfai_integrator.integrate1d
        q, ix = integrate(img, 2048, method='csr', unit='q_A^-1',
                          correctSolidAngle=True,
                          polarization_factor=0.999)

        dxrd = xrd1d(label=label, x=q, I=ix, xtype='q',
                     wavelength=self.wavelength)
        self.add_data(dxrd, label=label)


    def onCIFBrowse(self, event=None):
        shown = False
        if self.cif_browser is not None:
            try:
                self.cif_browser.Raise()
                shown = True
            except:
                del self.cif_browser
                shown = False
        if not shown:
            self.cif_browser = CIFFrame(usecif_callback=self.onLoadCIF,
                                        _larch=self.larch)
            self.cif_browser.Raise()

    def onLoadCIF(self, cif=None):
        if cif is None:
            return
        t0 = time.time()

        energy = E_from_lambda(self.wavelength)

        sfact = cif.get_structure_factors(wavelength=self.wavelength)
        try:
            self.cif_browser.cifdb.set_hkls(self.current_cif.ams_id, sfact.hkls)
        except:
            pass

        mineral = getattr(cif, 'mineral', None)
        label = getattr(mineral, 'name', '')
        if len(label) < 0:
            label = getattr(cif, 'formula', '')
        cifid = getattr(cif, 'ams_id', '')
        if len(label) < 1 and len(cifid) > 0:
            label = 'CIF:{cifid}'
        else:
            label = f'{label}, CIF:{cifid}'

        try:
            q = self.datasets[self.current_label].q
        except:
            q = np.linspace(0, 10, 2048)

        sigma = 2.5*(q[1] - q[0])

        intensity = q*0.0
        for cen, amp in zip(sfact.q, sfact.intensity):
            intensity += gaussian(q, amplitude=amp, center=cen, sigma=sigma)

        xdat = xrd1d(label=label, energy=energy, wavelength=self.wavelength)
        xdat.set_xy_data(np.array([q, intensity/max(intensity)]), 'q')
        self.add_data(xdat, label=label)


    def onSaveXY(self, event=None):
        fname = fix_filename(self.current_label.replace('.', '_') + '.xy')
        sfile = FileSave(self, 'Save XY Data', default_file=fname)
        if sfile is None:
            return

        label = self.current_label
        dset = self.datasets[label]

        xscale = self.wids['xscale'].GetSelection()
        xlabel = self.wids['xscale'].GetStringSelection()
        xdat = dset.twth
        xlabel = pyFAI.units.TTH_DEG
        if xscale == 0:
            xdat = dset.q
            xlabel = pyFAI.units.Q_A

        ydat = 1.0*dset.I/dset.scale
        wavelength = PLANCK_HC/(dset.energy*1000.0)
        buff = [f"# XY data from {label}",
                f"# wavelength (Ang) = {gformat(wavelength)}",
                "# Calibration data from pyFAI:"]
        for key, value in self.poni.items():
            buff.append(f"#    {key}: {value}")
        buff.append("#-------------------------------------")
        buff.append(f"# {xlabel}    Intensity")

        for x, y in zip(xdat, ydat):
            buff.append(f"  {gformat(x, 13)}  {gformat(y, 13)}")
        buff.append('')
        with open(sfile, 'w') as fh:
            fh.write('\n'.join(buff))


    def build(self):
        sizer = wx.GridBagSizer(3, 3)
        sizer.SetVGap(3)
        sizer.SetHGap(3)

        splitter = wx.SplitterWindow(self, style=wx.SP_LIVE_UPDATE)
        splitter.SetMinimumPaneSize(220)

        # left side: list of XRD 1D patterns
        lpanel = wx.Panel(splitter)
        lpanel.SetMinSize((275, 350))
        # rpanel = scrolled.ScrolledPanel(splitter)
        rpanel = wx.Panel(splitter)
        rpanel.SetMinSize((400, 350))
        rpanel.SetSize((750, 550))

        ltop = wx.Panel(lpanel)

        def Btn(msg, x, act):
            b = Button(ltop, msg, size=(x, 30),  action=act)
            b.SetFont(Font(FONTSIZE))
            return b

        sel_none = Btn('Select None', 130, self.onSelNone)
        sel_all  = Btn('Select All', 130, self.onSelAll)

        self.filelist = FileCheckList(lpanel, main=self,
                                      select_action=self.show_dataset,
                                      remove_action=self.remove_dataset)
        set_color(self.filelist, 'list_fg', bg='list_bg')

        tsizer = wx.BoxSizer(wx.HORIZONTAL)
        tsizer.Add(sel_all, 1, LEFT|wx.GROW, 1)
        tsizer.Add(sel_none, 1, LEFT|wx.GROW, 1)
        pack(ltop, tsizer)

        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add(ltop, 0, LEFT|wx.GROW, 1)
        sizer.Add(self.filelist, 1, LEFT|wx.GROW|wx.ALL, 1)
        pack(lpanel, sizer)

        # right side: parameters controlling display
        panel = GridPanel(rpanel, ncols=6, nrows=10, pad=3, itemstyle=LEFT)
        panel.sizer.SetVGap(3)
        panel.sizer.SetHGap(3)

        self.font_fixedwidth = wx.Font(FONTSIZE_FW, wx.MODERN, wx.NORMAL, wx.BOLD)

        # title row
        self.wids = wids = {}
        title = SimpleText(panel, '1D XRD Data Display', font=Font(FONTSIZE+2),
                           colour=COLORS['title'], style=LEFT)

        self.last_plot_type = 'one'
        self.plotone = Button(panel, 'Plot Current ', size=(125, -1),
                              action=self.onPlotOne)
        self.plotsel = Button(panel, 'Plot Selected ', size=(125, -1),
                              action=self.onPlotSel)
        wids['plotone'] = Choice(panel, choices=PLOT_CHOICES, default=0,
                                 action=self.onPlotOne, size=(200, -1))
        wids['plotsel'] = Choice(panel, choices=PLOT_CHOICES_MULTI, default=0,
                                 action=self.onPlotSel, size=(200, -1))
        wids['xscale'] = Choice(panel, choices=X_SCALES, default=0,
                                 action=self.onPlotEither, size=(100, -1))

        opts = dict(default=False, size=(200, -1), action=self.onPlotEither)
        wids['plot_win']  = Choice(panel, size=(100, -1), choices=PlotWindowChoices,
                                   action=self.onPlotEither)
        wids['plot_win'].SetStringSelection('1')

        wids['auto_scale'] = Check(panel, default=True, label='auto?',
                                   action=self.auto_scale)
        wids['scale_method'] = Choice(panel, choices=list(SCALE_METHODS.keys()),
                                      size=(250, -1), action=self.auto_scale, default=0)

        wids['scale'] = FloatCtrl(panel, value=1.0, size=(90, -1), precision=2,
                                  action=self.set_scale)
        wids['wavelength'] = SimpleText(panel, label="%.6f" % (self.wavelength), size=(100, -1))
        wids['energy_ev'] = SimpleText(panel, label="%.1f" % (PLANCK_HC/self.wavelength), size=(100, -1))

        wids['bkg_qwid'] = FloatSpin(panel, value=0.1, size=(90, -1), digits=2,
                                     increment=0.01,
                                     min_val=0.001, max_val=5, action=self.on_bkg)
        wids['bkg_nsmooth'] = FloatSpin(panel, value=30, size=(90, -1),
                                        digits=0, min_val=2, max_val=100, action=self.on_bkg)
        wids['bkg_porder'] = FloatSpin(panel, value=40, size=(90, -1),
                                        digits=0, min_val=2, max_val=100, action=self.on_bkg)

        def CopyBtn(name):
            return Button(panel, 'Copy to Seleceted', size=(150, -1),
                          action=partial(self.onCopyAttr, name))

        wids['bkg_copy'] = CopyBtn('bkg')
        wids['scale_copy'] = CopyBtn('scale_method')

        def slabel(txt):
            return wx.StaticText(panel, label=txt)

        panel.Add(title, style=LEFT, dcol=5)
        panel.Add(self.plotsel, newrow=True)
        panel.Add(wids['plotsel'], dcol=2)
        panel.Add(slabel(' X scale: '), dcol=2, style=LEFT)
        panel.Add(wids['xscale'], style=RIGHT)

        panel.Add(self.plotone, newrow=True)
        panel.Add(wids['plotone'], dcol=2)
        panel.Add(slabel(' Plot Window: '), dcol=2)
        panel.Add(wids['plot_win'], style=RIGHT)

        panel.Add((5, 5))
        panel.Add(HLine(panel, size=(550, 3)), dcol=6, newrow=True)
        panel.Add((5, 5))


        panel.Add(slabel(' Scaling Factor: '), style=LEFT, newrow=True)
        panel.Add(wids['scale'])
        panel.Add(wids['auto_scale'])
        panel.Add(slabel(' Scaling Method: '), style=LEFT, newrow=True)
        panel.Add(wids['scale_method'], dcol=3)
        panel.Add(wids['scale_copy'], dcol=2, style=RIGHT)

        panel.Add((5, 5))
        panel.Add(HLine(panel, size=(550, 3)), dcol=6, newrow=True)
        panel.Add((5, 5))

        panel.Add(slabel(' Background Subtraction Parameters: '), dcol=3, style=LEFT, newrow=True)
        panel.Add(wids['bkg_copy'], dcol=3, style=RIGHT)

        panel.Add(slabel(' Q width (\u212B\u207B\u00B9): '), style=LEFT, newrow=True)
        panel.Add(wids['bkg_qwid'])
        panel.Add(slabel(' Smoothing Steps: '), style=LEFT, newrow=True)
        panel.Add(wids['bkg_nsmooth'], dcol=2)
        panel.Add(slabel(' Polynomial Order: '), style=LEFT, newrow=True)
        panel.Add(wids['bkg_porder'])

        panel.Add((5, 5))
        panel.Add(HLine(panel, size=(550, 3)), dcol=6, newrow=True)
        panel.Add((5, 5))

        panel.Add(slabel(' Calibration, Calculated XRD Patterns: '), dcol=6, style=LEFT, newrow=True)
        panel.Add(slabel(' X-ray Energy (eV): '), style=LEFT, newrow=True)
        panel.Add(wids['energy_ev'], dcol=1)
        panel.Add(slabel(' Wavelength (\u212B): '), style=LEFT, newrow=False)
        panel.Add(wids['wavelength'], dcol=2)


        panel.Add((5, 5))

        panel.pack()

        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add((5, 5), 0, LEFT, 3)
        sizer.Add(panel, 0, LEFT, 3)
        sizer.Add((5, 5), 0, LEFT, 3)
        pack(rpanel, sizer)

        # rpanel.SetupScrolling()

        splitter.SplitVertically(lpanel, rpanel, 1)
        mainsizer = wx.BoxSizer(wx.VERTICAL)
        mainsizer.Add(splitter, 1, wx.GROW|wx.ALL, 5)
        pack(self, mainsizer)
        self.SetSize((875, 450))

        self.Show()
        self.Raise()

    def set_ponifile(self, ponifile, with_pyfai=True):
        "set poni from datafile"
        try:
            self.set_poni(read_poni(ponifile), with_pyfai=with_pyfai)
        except:
            pass

    def set_poni(self, poni, with_pyfai=True):
        "set poni from dict"
        try:
            self.poni.update(poni)
            self.set_wavelength(self.poni['wavelength']*1.e10)
            self.tiff_reader.Enable(self.poni.get('dist', -1) > 0)
        except:
            pass

        if with_pyfai:
            try:
                self.pyfai_integrator = AzimuthalIntegrator(**self.poni)
            except:
                self.pyfai_integrator = None
                

    def set_wavelength(self, value):
        self.wavelength = value
        self.wids['wavelength'].SetLabel("%.6f" % value)
        self.wids['energy_ev'].SetLabel("%.1f" % (PLANCK_HC/value))
        for key, dset in self.datasets.items():
            dset.set_wavelength(value)

    def onCopyAttr(self, name=None, event=None):
        # print("Copy ", name, event)
        if name == 'bkg':
            qwid = self.wids['bkg_qwid'].GetValue()
            nsmooth = int(self.wids['bkg_nsmooth'].GetValue())
            cheb_order = int(self.wids['bkg_porder'].GetValue())

            for label in self.filelist.GetCheckedStrings():
                dset = self.datasets.get(label, None)
                if dset is not None:
                    dset.bkg_qwid = qwid
                    dset.bkg_nsmooth = nsmooth
                    dset.bkg_porder = cheb_order
                    # print("redo bkg for ", label)
                    dset.bkgd = calc_bgr(dset, qwid=qwid, nsmooth=nsmooth,
                                         cheb_order=cheb_order)

        elif name == 'scale_method':
            for label in self.filelist.GetCheckedStrings():
                dset = self.datasets.get(label, None)
                if dset is not None:
                    # print("redo scale for ", label)
                    self.scale_data(dset, with_plot=False)


    def onSelNone(self, event=None):
        self.filelist.select_none()

    def onSelAll(self, event=None):
        self.filelist.select_all()

    def on_bkg(self, event=None, value=None):
        try:
            qwid = self.wids['bkg_qwid'].GetValue()
            nsmooth = int(self.wids['bkg_nsmooth'].GetValue())
            cheb_order = int(self.wids['bkg_porder'].GetValue())
        except:
            return
        label = self.current_label
        if label not in self.datasets:
            return
        dset = self.datasets[label]
        dset.bkgd = calc_bgr(dset, qwid=qwid, nsmooth=nsmooth,
                             cheb_order=cheb_order)
        if 'back' not in self.wids['plotone'].GetStringSelection().lower():
            self.wids['plotone'].SetSelection(1)
        else:
            self.onPlotOne()

    def show_dataset(self, event=None, label=None):
        # print('show xd1d ', event, label)
        if label is None and event is not None:
            label = str(event.GetString())
        if label not in self.datasets:
            return

        self.current_label = label
        dset = self.datasets[label]

        if not hasattr(dset, 'scale'):
            dset.scale = dset.I.max()
            dset.scale_method = 'raw_max'
            dset.auto_scale = True
            dset.bkg_qwid = 0.1
            dset.bkg_nsmooth = 30
            dset.bkg_porder = 40

        bkgd = getattr(dset, 'bkgd', None)
        if (bkgd is None
            or (isinstance(bkgd, np.ndarray)
                and (bkgd.sum() < 0.5/len(bkgd)))):
            dset.bkgd = calc_bgr(dset)

        meth_desc = keyof(SCALE_METHODS, dset.scale_method, 'raw_max')

        self.wids['scale_method'].SetStringSelection(meth_desc)
        self.wids['auto_scale'].SetValue(dset.auto_scale)
        self.wids['scale'].SetValue(dset.scale)
        self.wids['energy_ev'].SetLabel("%.1f" % (dset.energy*1000.0))
        self.wids['wavelength'].SetLabel("%.6f" % (PLANCK_HC/(dset.energy*1000.0)))

        self.wids['bkg_qwid'].SetValue(dset.bkg_qwid)
        self.wids['bkg_nsmooth'].SetValue(dset.bkg_nsmooth)
        self.wids['bkg_porder'].SetValue(dset.bkg_porder)

        self.onPlotOne(label=label)

    def set_scale(self, event=None, value=-1.0):
        label = self.current_label
        if label not in self.datasets:
            return
        if value < 0:
            value = self.wids['scale'].GetValue()
        self.datasets[label].scale = value # self.wids['scale'].GetValue()

    def auto_scale(self, event=None):
        label = self.current_label
        if label not in self.datasets:
            return
        dset = self.datasets[label]
        dset.auto_scale = self.wids['auto_scale'].IsChecked()
        self.wids['scale_method'].Enable(dset.auto_scale)

        if dset.auto_scale:
            self.scale_data(dset, with_plot=True)

    def scale_data(self, dset, with_plot=True):
        meth_name = self.wids['scale_method'].GetStringSelection()

        meth = dset.scale_method = SCALE_METHODS[meth_name]

        # if not meth.startswith('raw'):
        qwid = self.wids['bkg_qwid'].GetValue()
        nsmooth = int(self.wids['bkg_nsmooth'].GetValue())
        cheb_order = int(self.wids['bkg_porder'].GetValue())
        dset.bkgd = calc_bgr(dset, qwid=qwid, nsmooth=nsmooth,
                            cheb_order=cheb_order)
        dset.bkg_qwid = qwid
        dset.bkg_nmsooth = nsmooth
        dset.bkg_porder = cheb_order

        scale =  -1
        if meth == 'raw_max':
            scale = dset.I.max()
        elif meth == 'raw_mean':
            scale = dset.I.mean()
        elif meth == 'sub_max':
            scale = (dset.I - dset.bkgd).max()
        elif meth == 'sub_mean':
            scale = (dset.I - dset.bkgd).mean()
        elif meth == 'bkg_max':
            scale = (dset.bkgd).max()
        elif meth == 'bkg_mean':
            scale = (dset.bkgd).mean()

        if scale > 0:
            self.wids['scale'].SetValue(scale)
            if with_plot:
                self.onPlotOne()

    def rename_dataset(self, newlabel):
        dset = self.datasets.pop(self.current_label)
        dset.label = newlabel
        self.datasets[newlabel] = dset
        self.current_label = newlabel

        self.filelist.Clear()
        for name in self.datasets:
            self.filelist.Append(name)


    def onRenameDataset(self, event=None):
        RenameDialog(self, self.current_label, self.rename_dataset).Show()


    def remove_dataset(self, dname=None, event=None):
        if dname in self.datasets:
            self.datasets.pop(dname)

        self.filelist.Clear()
        for name in self.datasets:
            self.filelist.Append(name)


    def remove_selected_datasets(self, event=None):
        sel = []
        for checked in self.filelist.GetCheckedStrings():
            sel.append(str(checked))
        if len(sel) < 1:
            return

        dlg = RemoveDialog(self, sel)
        res = dlg.GetResponse()
        dlg.Destroy()

        if res.ok:
            all = self.filelist.GetItems()
            for dname in sel:
                self.datasets.pop(dname)
                all.remove(dname)

            self.filelist.Clear()
            for name in all:
                self.filelist.Append(name)

    def get_imdisplay(self, win=1):
        wintitle='XRD Image Window %i' % win
        opts = dict(wintitle=wintitle, win=win, image=True)
        self.img_display = self.larch.symtable._plotter.get_display(**opts)
        return self.img_display

    def get_display(self, win=1, stacked=False):
        wintitle='XRD Plot Window %i' % win
        opts = dict(wintitle=wintitle, stacked=stacked, win=win, linewidth=3)
        self.plot_display = self.larch.symtable._plotter.get_display(**opts)
        return self.plot_display

    def plot_dset(self, dset, plottype, newplot=True):
        win    = int(self.wids['plot_win'].GetStringSelection())
        xscale = self.wids['xscale'].GetSelection()
        opts = {'show_legend': True, 'xmax': None,
                'xlabel':  self.wids['xscale'].GetStringSelection(),
                'ylabel':'Scaled Intensity',
                'label': dset.label}

        xdat = dset.q
        if xscale == 2:
           xdat = dset.d
           opts['xmax'] = min(12.0, max(xdat))
        elif xscale == 1:
            xdat = dset.twth

        ydat = 1.0*dset.I/dset.scale
        if plottype == 'sub':
            ydat = 1.0*(dset.I-dset.bkgd)/dset.scale
            opts['ylabel'] = 'Scaled (Intensity - Background)'

        pframe = self.get_display(win=win)
        plot = pframe.plot if newplot else pframe.oplot
        plot(xdat, ydat, **opts)
        if plottype == 'raw+bkg':
            y2dat = 1.0*dset.bkgd/dset.scale
            opts['ylabel'] = 'Scaled Intensity with Background'
            opts['label'] = 'background'
            pframe.oplot(xdat, y2dat, **opts)

    def onPlotOne(self, event=None, label=None):
        if label is None:
            label = self.current_label
        if label not in self.datasets:
            return
        dset = self.datasets[label]
        self.last_plot_type = 'one'
        plottype = PLOT_TYPES.get(self.wids['plotone'].GetStringSelection(), 'raw')
        self.plot_dset(dset, plottype, newplot=True)
        wx.CallAfter(self.SetFocus)

    def onPlotSel(self, event=None):
        labels = self.filelist.GetCheckedStrings()
        if len(labels) < 1:
            return
        self.last_plot_type = 'multi'
        plottype = PLOT_TYPES.get(self.wids['plotsel'].GetStringSelection(), 'raw')
        newplot = True
        for label in labels:
            dset = self.datasets.get(label, None)
            if dset is not None:
                self.plot_dset(dset, plottype, newplot=newplot)
                newplot = False
        wx.CallAfter(self.SetFocus)

    def onPlotEither(self, event=None):
        if self.last_plot_type == 'multi':
            self.onPlotSel(event=event)
        else:
            self.onPlotOne(event=event)

    def add_data(self, dataset, label=None,  **kws):
        if label is None:
            label = 'XRD pattern'
        if label in self.datasets:
            print('label already in datasets: ', label )
        else:
            self.filelist.Append(label)
            self.datasets[label] = dataset
            self.show_dataset(label=label)

class XRD1DApp(LarchWxApp):
    def __init__(self, **kws):
        LarchWxApp.__init__(self)

    def createApp(self):
        frame = XRD1DFrame()
        frame.Show()
        self.SetTopWindow(frame)
        return True
